/******************************************************************
 
        MUSCLE SmartCard Development ( http://www.musclecard.com )
            Title  : authentIC.c
            Package: authentIC
            Author : David Corcoran
            Date   : 02/19/02
            License: Copyright (C) 2002 David Corcoran
                     <corcoran@linuxnet.com>
            Purpose: A DDK for MuscleCard plug-ins.
	             
 
********************************************************************/

/********************************************************
PlugIn for Oberthur AuthentIC with applet X509

Copyright (C) 2002 IDEALX (http://www.idealx.com/).

This program is free software; you can redistribute it and/or
modify it under the terms of either the GNU General Public License
or the Clarified Artistic License, as specified in the LICENSE
file.

author: Victor Tarasov vtarasov@idealx.com
********************************************************/

#ifdef MSC_TARGET_OSX
#include <PCSC/musclecard.h>
#else
#include <musclecard.h>
#include <debuglog.h>
#endif

#include "pl_musclecard.h"
#include "authentIC.h"
#include <string.h>
#include <stdio.h>
#include <errno.h>

#ifdef MSC_ENABLE_DMALLOC
#include <dmalloc.h>
#endif

//#define DEBUGLOG_STDOUT

#ifdef MSC_DEBUG
#   ifdef DEBUGLOG_STDOUT
#       define DPRINTF(format,args...) printf(format,## args)
#   else
#       define DPRINTF(format,args...) autentIC_Log(format, ## args)
#   endif
#else
#   define DPRINTF(format,args...)
#endif

MSCUChar8 privateKeyACL[] = { 0x4F, 0x4F, 0xFF, 0x4F, 0xFF, 0xFF, 0xFF, 0xFF};
MSCUChar8 publicKeyACL[] = { 0x4F, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF};
    
typedef struct  {
    char *name;
    short int dir[5];
    short int id;
    unsigned char type;
} MSCObject_t;

MSCUChar8 authentICPins[2] = {SO_PIN, USER_PIN};

#define NN_OBJECTS 3

MSCObject_t objects[NN_OBJECTS] = {
    {"ObjectInfo", 
        { AUTHENTIC_DIR_ID, 0, 0, 0, 0},
        AUTHENTIC_OBJECTINFO_ID,
        FILE_TYPE_EF_VARIABLE
    },
    
    {"UserInfo", 
        { AUTHENTIC_DIR_ID, 0, 0, 0, 0},
        AUTHENTIC_USERINFO_ID,
        FILE_TYPE_EF_TRANSPARENT
    },
    
    {"KeyInfo", 
        { AUTHENTIC_DIR_ID, 0, 0, 0, 0},
        AUTHENTIC_KEYINFO_ID,
        FILE_TYPE_EF_TRANSPARENT
    },
};

/* Keep the name of the previously selected object. */
static MSCUChar8 lastSelectedName[MSC_MAXSIZE_OBJID];

#define autentIC_LogFilename "/tmp/autentIC.log"

#define AUTENTHIC_COMMON_COMMAND_RETURN                     \
    if ( tBuffer.apduResponseSize < 2 )                     \
        return MSC_UNSPECIFIED_ERROR;                       \
    if (tBuffer.apduResponse[tBuffer.apduResponseSize - 2]  \
            == AUMSC_SUCCESS_BYTE)                          \
        return MSC_SUCCESS;                                 \
    else                                                    \
        return convertSW(tBuffer.apduResponse +             \
                tBuffer.apduResponseSize - 2);         

#define SELECT_MF                                               \
{                                                               \
    do {                                                        \
        rv = auSelectFile( pConnection, 0x3F00,             \
                 SELECT_MODE_PARENT|SELECT_MODE_WITH_ANSWER,    \
                 response, &responseLen);                       \
    } while (rv==MSC_SUCCESS && responseLen > 2 &&              \
            !(response[4]==0x38 && response[8]==0x3F &&         \
            response[9]==0x00));                                \
    if ( rv != MSC_SUCCESS ) return rv;                         \
}

#define SELECT_AUTHENTIC_DIR                                    \
    SELECT_MF;                                                  \
    rv = auSelectFile( pConnection, AUTHENTIC_DIR_ID,       \
            SELECT_MODE_DF|SELECT_MODE_WITH_ANSWER,             \
            response, &responseLen);                            \
    if ( rv != MSC_SUCCESS ) return rv;
    

#define MSC_RETURN_IF_ERROR(x)              \
do {                                        \
    rv = (x);                               \
    if (rv != MSC_SUCCESS)  {               \
        DPRINTF("%s(%i): MSC_ERROR: %X\n",  \
                    __FILE__,__LINE__,rv);  \
        return rv;                          \
    }                                       \
}while(0)

#define SCARD_RETURN_IF_ERROR(x)                \
do {                                            \
    rv = (x);                                   \
    if (rv != SCARD_S_SUCCESS)  {               \
        DPRINTF("%s(%i): SCARD_ERROR: %X\n",    \
                __FILE__,__LINE__,rv);          \
        return  pcscToMSC(rv);                  \
    }                                           \
}while(0)

static MSCUChar8 keyInfoBytes[AUTHENTIC_INFOBJ_LEN];

typedef struct {
    MSCUChar8  pBuffer[MAX_BUFFER_SIZE];
    MSCULong32 bufferSize;
    MSCUChar8  apduResponse[MAX_BUFFER_SIZE];
    MSCULong32 apduResponseSize;
    LPSCARD_IO_REQUEST ioType;
}   MSCTransmitBuffer, *MSCLPTransmitBuffer;


struct FileInfo {
        char *path;
        MSCUChar8 type;
        MSCUShort16 id, size;
        struct FileInfo *list;
        struct FileInfo *next;
        struct FileInfo *prev;
};

typedef struct FileInfo tFileInfo;

/* internal function */
void  autentIC_Log(char *args, ...);
void getOberthurACL(MSCPUChar8 out_acl, MSCPUChar8 in_acl);

MSCLong32 auSCardExchangeAPDU( MSCLPTokenConnection, MSCLPTransmitBuffer);

MSCLong32 auSCardSendAPDU( MSCLPTokenConnection, MSCLPTransmitBuffer);

MSC_RV auSelectFile(MSCLPTokenConnection pConnection,  
        MSCULong32 fileID, MSCUChar8 flag, 
        MSCPUChar8 apduResponse, MSCPUShort16 lenResponse); 

MSC_RV auCreateFile( MSCLPTokenConnection pConnection, MSCUChar8 fileType,
        MSCShort16 objectID,	MSCULong32 objectSize, 
        MSCPUChar8 pObjectACL ); 

MSC_RV auUpdateBinaryEF( MSCLPTokenConnection pConnection, 
        MSCULong32 offset, MSCPUChar8 pInputData, MSCUChar8 dataSize );

MSC_RV auReadBinary( MSCLPTokenConnection pConnection, 
		       MSCULong32 offset, MSCPUChar8 pOutputData, MSCUChar8 dataSize );

MSC_RV auDeleteFile( MSCLPTokenConnection pConnectioin,
             MSCUShort16 fileID, MSCUChar8 fileType);

MSC_RV pcscToMSC( MSCLong32 );

MSCUShort16 convertSW(MSCPUChar8 pBuffer);

void INSPrintDebug(MSCUChar8 ch);

MSC_RV getObjectID( MSCLPTokenConnection pConnection, MSCPUShort16 objectID, 
        MSCString objectString, MSCUShort16 valid_name); 


MSC_RV auReadRecord( MSCLPTokenConnection pConnection, 
		       MSCUChar8 index, MSCUChar8 size, 
               MSCPUChar8 pOutputData, MSCPUShort16 sizeRecord );

MSC_RV auDeleteRecord( MSCLPTokenConnection pConnection, MSCUChar8 index);

MSC_RV auAppendRecord( MSCLPTokenConnection pConnection,
		       MSCPUChar8 pInputData, MSCUChar8 dataSize );


MSC_RV auGenerateKeyPair( MSCLPTokenConnection pConnection, MSCUShort16 prvKey,
			MSCUShort16 pubKey, MSCUChar8 keySize ); 

MSC_RV auDeleteKeyInfo( MSCLPTokenConnection pConnection, MSCUChar8 keyNum);

MSC_RV auReadKeyInfo( MSCLPTokenConnection pConnection, MSCUChar8 keyNum,  
        MSCLPKeyInfo pKeyInfo);

MSC_RV auListFiles( MSCLPTokenConnection pConnection, MSCUChar8 seqOption, 
                        MSCPUShort16 fileID);

MSC_RV auAppendObjectInfo( MSCLPTokenConnection pConnection, 
        MSCLPObjectInfo pObjectInfo);

MSC_RV auDeleteObjectInfo( MSCLPTokenConnection pConnection, MSCString objectID);

MSC_RV auCleanFramework( MSCLPTokenConnection pConnection);

MSC_RV auWriteKeyInfo( MSCLPTokenConnection pConnection, 
        MSCUChar8 keyNum,  MSCUChar8 keyType, MSCUChar8 keyPartner, 
        MSCUShort16 keySize, MSCLPKeyPolicy keyPolicy, 
        MSCLPKeyACL keyACL);

MSC_RV auReadPublicKey( MSCLPTokenConnection pConnection, MSCUShort16 id,
		       MSCPUChar8 pExp, MSCPUChar8 pMod);

MSC_RV auUpdateKeyComponent( MSCLPTokenConnection pConnection, 
		       key_component_t compType, MSCPUChar8 compData, MSCUShort16 compLen); 

MSC_RV auEncrypt( MSCLPTokenConnection pConnection, MSCUChar8 key_num,
			MSCPUChar8 pInputData, MSCULong32 inputDataSize, 
            MSCPUChar8 pOutputData, MSCPULong32 outputDataSize);

MSC_RV auEncryptPublic( MSCLPTokenConnection pConnection, MSCUChar8 keyNum,
			MSCPUChar8 pInputData, MSCULong32 inputDataSize, 
            MSCPUChar8 pOutputData, MSCPULong32 outputDataSize); 

MSC_RV auDecrypt( MSCLPTokenConnection pConnection, MSCUChar8 key_num,
			MSCPUChar8 pInputData, MSCULong32 inputDataSize, 
            MSCPUChar8 pOutputData, MSCPULong32 outputDataSize);

MSC_RV auDecryptPublic( MSCLPTokenConnection pConnection, MSCUChar8 key_num,
			MSCPUChar8 pInputData, MSCULong32 inputDataSize, 
            MSCPUChar8 pOutputData, MSCPULong32 outputDataSize);

MSC_RV auGetPublicRSAKey(MSCLPTokenConnection pConnection, 
            MSCUChar8 keyNum, RSA **rsaOutput);

MSC_RV auImportRSAKey( MSCLPTokenConnection pConnection, MSCUChar8 keyNum,
		    MSCLPKeyACL pKeyACL, RSA *rsa, MSCLPKeyPolicy keyPolicy,
		    MSCUChar8 keyType, MSCUChar8 keyPartner );

MSC_RV auUnblockPIN( MSCLPTokenConnection pConnection, 
            MSCUChar8 pinNumber,
		    MSCPUChar8 pUnblockCode, MSCULong32 unblockCodeSize,
            MSCPUChar8 pNewUserPIN, MSCULong32 newUserPINSize);

void MemCopyTo16(MSCPUShort16 destValue, void *value); 
void MemCopy16(void *outValue, MSCPUShort16 srcValue); 


MSC_RV auListFilesAllocate( MSCLPTokenConnection pConnection,
        MSCPUChar8 path, struct FileInfo **retList);


/********** musclecard-framework functions *********************************/
/*
 * Create smart card authentIC framework.
 */
MSC_RV PL_MSCWriteFramework( MSCLPTokenConnection pConnection,
			  MSCLPInitTokenParams pInitParams ) 
{
    MSCLong32 rv;
    MSCUChar8 dirACL[] = { 0x4F, 0x4F, 0xFF, 0x4F, 0xFF, 0xFF, 0xFF, 0xFF};
    MSCUChar8 publicInfoACL[] = { 0x4F, 0x4F, 0x0F, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF};
    MSCUChar8 publicObjectACL[] = { 0x4F, 0x4F, 0x0F, 0x4F, 0xFF, 0xFF, 0xFF, 0xFF};
    MSCUChar8 response[MSC_MAXSIZE_BUFFER];
    MSCUShort16 responseLen;
    int ii;
    MSCPUChar8 defaultPIN = "Muscle00";
    MSCPUChar8 defaultSOPIN = "Muscle01";
    
    MSCUChar8 pin[15];
    MSCUChar8  pinLen = 0;

    MSCUChar8 sopin[15];
    MSCUChar8 sopinLen = 0;
    
    DPRINTF("%s(%i): MSCWriteFramework\n",__FILE__,__LINE__);
    
    strncpy(pin, defaultPIN, 15);
    pinLen = strlen(defaultPIN) > 15 ? 15 : strlen(defaultPIN);
    
    strncpy(sopin, defaultSOPIN, 15);
    sopinLen = strlen(defaultSOPIN) > 15 ? 15 : strlen(defaultSOPIN);
    
#if 1
    if ( pInitParams->transportBehavior == MSC_INIT_DEFAULT_KEY ) {
        /* Do nothing */
    } else if ( pInitParams->transportBehavior == MSC_INIT_USE_KEY ) {
        
        if ( pInitParams->transportKeyLen < 4 ||  
                pInitParams->transportKeyLen > 15) {
            return MSC_INVALID_PARAMETER;
        }
        memcpy(sopin, pInitParams->transportKey, pInitParams->transportKeyLen);
        sopinLen = pInitParams->transportKeyLen;
    }

// When creating DF, Obertur do not need to know it's largest possible total size in bytes.
// So,  pInitParams->objectMemory is not used here.
    if ( pInitParams->objectMemory == 0 ) {
        pInitParams->objectMemory = 8500;
    } 
#endif

    DPRINTF("%s(%i): MSCWriteFramework: sopin %s; len %i\n",
            __FILE__,__LINE__,sopin, sopinLen);
    DPRINTF("%s(%i): MSCWriteFramework: pin %s; len %i\n",
            __FILE__,__LINE__,pin, pinLen);

    if ( pInitParams->transportBehavior != MSC_INIT_IGNORE_KEY ) {
        MSC_RETURN_IF_ERROR(PL_MSCVerifyPIN( pConnection, NUM_SOPIN, 
                    &sopin[0], sopinLen));
    }
    
    auCleanFramework(pConnection);

    SELECT_MF;    
    
    /* Root authentIC directory. */
    MSC_RETURN_IF_ERROR( auCreateFile(pConnection, 
                FILE_TYPE_DF, AUTHENTIC_DIR_ID, 0, dirACL));
   
    MSC_RETURN_IF_ERROR( auSelectFile( pConnection, 
                AUTHENTIC_DIR_ID, SELECT_MODE_DF|SELECT_MODE_WITH_ANSWER,
                response, &responseLen));
    
    /* Objects info. */
    MSC_RETURN_IF_ERROR( auCreateFile(pConnection, FILE_TYPE_EF_VARIABLE,
                AUTHENTIC_OBJECTINFO_ID,  MAX_NUMBER_OBJECT_INFO, publicObjectACL));
   
    /* KeyInfo object */
    MSC_RETURN_IF_ERROR( auCreateFile(pConnection, FILE_TYPE_EF_VARIABLE,
                AUTHENTIC_KEYINFO_ID,  AUTHENTIC_MAX_KEYS, publicObjectACL));
   
    /* Container for the keys. */
    MSC_RETURN_IF_ERROR( auCreateFile(pConnection, FILE_TYPE_DF, 
                AUTHENTIC_KEYDIR_ID, 0, dirACL));

    /* Container for the objects. */
    MSC_RETURN_IF_ERROR( auCreateFile(pConnection, FILE_TYPE_DF, 
                AUTHENTIC_OBJECTDIR_ID, 0, dirACL));
    
    SELECT_MF;

    if (pInitParams->newTransportKey!=NULL && 
            pInitParams->newTransportKeyLen > 3 &&  pInitParams->newTransportKeyLen < 16)   {
        MSC_RETURN_IF_ERROR(  
                PL_MSCChangePIN( pConnection, NUM_SOPIN, sopin, sopinLen,
                    pInitParams->newTransportKey, pInitParams->newTransportKeyLen )
        );

        memcpy(sopin, pInitParams->newTransportKey, pInitParams->newTransportKeyLen);
        sopinLen = pInitParams->newTransportKeyLen;

        MSC_RETURN_IF_ERROR(PL_MSCVerifyPIN( pConnection, NUM_SOPIN, 
                    &sopin[0], sopinLen));
    }

    if (pInitParams->defaultCHV!=NULL && 
            pInitParams->defaultCHVLen > 3 && pInitParams->defaultCHVLen < 16)  {
        DPRINTF("%s(%i): MSCWriteFramework\n",__FILE__,__LINE__);
        MSC_RETURN_IF_ERROR(  
                auUnblockPIN( pConnection, NUM_PIN, sopin, sopinLen, 
                    pInitParams->defaultCHV, pInitParams->defaultCHVLen) 
        );
        
        MSC_RETURN_IF_ERROR(PL_MSCVerifyPIN( pConnection, NUM_PIN, 
                    pInitParams->defaultCHV, pInitParams->defaultCHVLen));
    }

    DPRINTF("%s(%i): MSCWriteFramework\n",__FILE__,__LINE__);
    return MSC_SUCCESS;
}


/*
 * Used to select authentIC x509 Smartcard Application.
 */
MSC_RV PL_MSCIdentifyToken( MSCLPTokenConnection pConnection) 
{
    MSCLong32 rv;
    MSCTransmitBuffer tBuffer;
    MSCPUChar8 apduResponse;
    MSCPUChar8 pBuffer;

    DPRINTF("%s(%i): PL_MSCIdentifyToken: size %d\n",
            __FILE__,__LINE__, pConnection->tokenInfo.tokenAppLen);

    pBuffer = tBuffer.pBuffer; 
    apduResponse = tBuffer.apduResponse;
	  
    pBuffer[OFFSET_CLA] = 0;
    pBuffer[OFFSET_INS] = 0xA4;
    pBuffer[OFFSET_P1]  = 0x04;
    pBuffer[OFFSET_P2]  = 0x00;
    pBuffer[OFFSET_LC]  = pConnection->tokenInfo.tokenAppLen;
 
    memcpy(&pBuffer[OFFSET_DATA], pConnection->tokenInfo.tokenApp,
            pConnection->tokenInfo.tokenAppLen);
    
    pBuffer[OFFSET_DATA + pConnection->tokenInfo.tokenAppLen] =  
            pConnection->tokenInfo.tokenAppLen + 4;
    tBuffer.bufferSize =  pConnection->tokenInfo.tokenAppLen + 6;
    tBuffer.apduResponseSize = MSC_MAXSIZE_BUFFER;

    SCARD_RETURN_IF_ERROR( auSCardExchangeAPDU(pConnection, &tBuffer));
    
    pConnection->loggedIDs = 0;

    if(tBuffer.apduResponseSize >= 2) {
        if(apduResponse[tBuffer.apduResponseSize-2] == AUMSC_SUCCESS_BYTE)    {
            return MSC_SUCCESS;
        }
        else 
            return convertSW(apduResponse + tBuffer.apduResponseSize-2 );
    } 
    else {
        return MSC_UNSPECIFIED_ERROR;
    }
}


MSC_RV PL_MSCInitializePlugin( MSCLPTokenConnection pConnection ) 
{
    return MSC_SUCCESS;
}

MSC_RV PL_MSCFinalizePlugin( MSCLPTokenConnection pConnection ) 
{
    return MSC_SUCCESS;
}


/* 
 * Token Info: card type and version, EEPROM size and free space;
 */
MSC_RV PL_MSCGetStatus( MSCLPTokenConnection pConnection, 
		     MSCLPStatusInfo pStatusInfo ) 
{
    MSCLong32 rv;
    MSCTransmitBuffer tBuffer;
    MSCPUChar8 apduResponse;
    MSCPUChar8 pBuffer;
    MSCObjectInfo objInfo;
    
    DPRINTF("%s(%i): MSCGetStatus\n",__FILE__,__LINE__);

    tBuffer.pBuffer[OFFSET_CLA] = CLA_80;
    tBuffer.pBuffer[OFFSET_INS] = INS_GET_DATA;
    tBuffer.pBuffer[OFFSET_P1]  = 0xDF;
    tBuffer.pBuffer[OFFSET_P2]  = 0x64;
    tBuffer.pBuffer[OFFSET_LC]  = 0x05;
    
    tBuffer.bufferSize = 5;
    tBuffer.apduResponseSize = MSC_MAXSIZE_BUFFER;
    
    SCARD_RETURN_IF_ERROR( auSCardSendAPDU(pConnection, &tBuffer));

    if ( tBuffer.apduResponseSize < 2 ) 
        return MSC_UNSPECIFIED_ERROR;
            
    if (tBuffer.apduResponseSize != 7)   
            return convertSW(tBuffer.apduResponse + tBuffer.apduResponseSize-2 );

    pStatusInfo->freeMemory  = tBuffer.apduResponse[3]*0x100 +  tBuffer.apduResponse[4];
    pStatusInfo->totalMemory = pStatusInfo->freeMemory;
    rv = PL_MSCListObjects( pConnection, MSC_SEQUENCE_RESET, &objInfo );

    while (rv == MSC_SUCCESS) {
        pStatusInfo->totalMemory += objInfo.objectSize;
        rv = PL_MSCListObjects( pConnection, MSC_SEQUENCE_NEXT, &objInfo );
    }

    tBuffer.pBuffer[OFFSET_P2]  = 0x65;
    tBuffer.apduResponseSize = MSC_MAXSIZE_BUFFER;
    
    SCARD_RETURN_IF_ERROR( auSCardSendAPDU(pConnection, &tBuffer));

    if ( tBuffer.apduResponseSize < 2 ) 
        return MSC_UNSPECIFIED_ERROR;
            
    if (tBuffer.apduResponseSize != 7)   
        return convertSW(tBuffer.apduResponse + tBuffer.apduResponseSize-2 );

    pStatusInfo->appVersion  = tBuffer.apduResponse[3];
    pStatusInfo->swVersion  = tBuffer.apduResponse[4];
    
    pStatusInfo->usedPINs    = 0x01;
    pStatusInfo->usedKeys    = 0x00;
    pStatusInfo->loggedID    = pConnection->loggedIDs;

    return MSC_SUCCESS;
}


MSC_RV PL_MSCGetCapabilities( MSCLPTokenConnection pConnection, MSCULong32 Tag,
			      MSCPUChar8 Value, MSCPULong32 Length ) 
{
    MSCULong32  ulValue;
    MSCUShort16 usValue;
    MSCUChar8  ucValue;
    MSCUChar8  tagType;

  /* 4 - MSCULong32, 2 - MSCUShort16, 1 - MSCUChar8 */

    DPRINTF("%s(%i):  MSCGetCapabilities\n",__FILE__,__LINE__);
    ulValue = 0; usValue = 0; ucValue = 0; tagType = 0;

    switch(Tag) {

    case MSC_TAG_SUPPORT_FUNCTIONS:
        ulValue = 
            MSC_SUPPORT_GENKEYS | 
            MSC_SUPPORT_IMPORTKEY |
            MSC_SUPPORT_EXPORTKEY | 
            MSC_SUPPORT_COMPUTECRYPT | 
//            MSC_SUPPORT_EXTAUTH | 
            MSC_SUPPORT_LISTKEYS |
//            MSC_SUPPORT_CREATEPIN |
            MSC_SUPPORT_VERIFYPIN | 
//            MSC_SUPPORT_CHANGEPIN | 
            MSC_SUPPORT_UNBLOCKPIN | 
            MSC_SUPPORT_LISTPINS | 
            MSC_SUPPORT_CREATEOBJECT | 
            MSC_SUPPORT_DELETEOBJECT | 
            MSC_SUPPORT_WRITEOBJECT | 
            MSC_SUPPORT_READOBJECT |
            MSC_SUPPORT_LISTOBJECTS | 
//            MSC_SUPPORT_LOGOUTALL |
            MSC_SUPPORT_GETCHALLENGE;
            tagType = 4;
        break;

    case MSC_TAG_SUPPORT_CRYPTOALG:
        ulValue = MSC_SUPPORT_RSA;
        tagType = 4;
        break;

    case MSC_TAG_CAPABLE_RSA:
        ulValue = 
            MSC_CAPABLE_RSA_1024 | 
//            MSC_CAPABLE_RSA_768 |
//            MSC_CAPABLE_RSA_NOPAD | 
            MSC_CAPABLE_RSA_KEYGEN |
            MSC_CAPABLE_RSA_PKCS1;
        tagType = 4;
        break;

    case MSC_TAG_CAPABLE_DSA:
//        ulValue = 
//            MSC_CAPABLE_RSA_1024 | 
//            MSC_CAPABLE_RSA_768 |
//            MSC_CAPABLE_DSA_KEYGEN;
        tagType = 4;
        break;

    case MSC_TAG_CAPABLE_DES:
//        ulValue = MSC_CAPABLE_DES_ECB;
        tagType = 4;
        break;

    case MSC_TAG_CAPABLE_3DES:
        ulValue = 0;
        tagType = 4;
        break;

    case MSC_TAG_CAPABLE_OBJ_ATTR:
        ulValue = MSC_CAPABLE_OBJ_ZERO;
        tagType = 4;
        break;

    case MSC_TAG_CAPABLE_OBJ_IDSIZE:
        ucValue = 4;
        tagType = 1;
        break;

    case MSC_TAG_CAPABLE_OBJ_AUTH:
        usValue = MSC_AUT_ALL;
        tagType = 2;
        break;

    case MSC_TAG_CAPABLE_OBJ_MAXNUM:
        ulValue = 100;
        tagType = 4;
        break;

    case MSC_TAG_CAPABLE_PIN_ATTR:
        ulValue = MSC_CAPABLE_PIN_LEAVE;
        tagType = 4;
        break;

    case MSC_TAG_CAPABLE_PIN_MAXNUM:
        ucValue = 2;
        tagType = 1;
        break;

    case MSC_TAG_CAPABLE_PIN_MINSIZE:
        ucValue = 4;
        tagType = 1;
        break;

    case MSC_TAG_CAPABLE_PIN_MAXSIZE:
        ucValue = 15;
        tagType = 1;
        break;
    
    case MSC_TAG_CAPABLE_PIN_CHARSET:
        ulValue = 
            MSC_CAPABLE_PIN_A_Z | MSC_CAPABLE_PIN_a_z | 
            MSC_CAPABLE_PIN_0_9 | MSC_CAPABLE_PIN_CALC | MSC_CAPABLE_PIN_NONALPHA;
        tagType = 4;
        break;

    case MSC_TAG_CAPABLE_PIN_POLICY:
        ulValue = 0;
        tagType = 4;
        break;

    case MSC_TAG_CAPABLE_ID_STATE:
        ucValue = MSC_CAPABLE_ID_STATE;
        tagType = 1;
        break;

    case MSC_TAG_CAPABLE_RANDOM_MAX:
        ucValue = 255;
        tagType = 1;
        break;

    case MSC_TAG_CAPABLE_RANDOM_MIN:
        ucValue = 8;
        tagType = 1;
        break;

    case MSC_TAG_CAPABLE_KEY_AUTH:
        usValue = MSC_AUT_PIN_0;
        tagType = 2;
        break;

    case MSC_TAG_CAPABLE_PIN_AUTH:
        usValue = MSC_AUT_ALL;
        tagType = 2;
        break;


    default:
        return MSC_INVALID_PARAMETER;
    }

    switch(tagType) {
        case 1:
            memcpy(Value, &ucValue, 1);
            break;
        case 2:
            memcpy(Value, &usValue, 2);
            break;
        case 4:
            memcpy(Value, &ulValue, 4);
            break;
    }

    *Length = tagType;

    return MSC_SUCCESS;
}



MSC_RV PL_MSCExtendedFeature( MSCLPTokenConnection pConnection, 
			   MSCULong32 extFeature,
			   MSCPUChar8 outData, MSCULong32 outLength, 
			   MSCPUChar8 inData, MSCPULong32 inLength ) 
{

  DPRINTF("%s(%i):MSCExtendedFeature \n",__FILE__,__LINE__);
  return MSC_UNSUPPORTED_FEATURE;
}


/*
 * Public/Private key generation.
 * It cleans KeyNum key files, if exists.
 */
MSC_RV PL_MSCGenerateKeys( MSCLPTokenConnection pConnection, MSCUChar8 prvKeyNum,
			MSCUChar8 pubKeyNum, MSCLPGenKeyParams pParams ) 
{
    MSCLong32 rv;
    MSCTransmitBuffer transmitBuffer;
    MSCPUChar8 apduResponse;
    MSCKeyACL privACL;
    MSCKeyACL pubACL;
    MSCUChar8 cryptAlgo;
    MSCPUChar8 pBuffer;
    
    MSCUChar8 response[MSC_MAXSIZE_BUFFER];
    MSCUShort16 responseLen;

    MSCUChar8 pExp[MSC_EXP_SIZE];
    MSCUChar8 pMod[MSC_MOD_SIZE];

    MSCUShort16 prvKeyID = prvKeyNum + AUTHENTIC_KEY_ID;
    MSCUShort16 pubKeyID = pubKeyNum + AUTHENTIC_KEY_ID;

    DPRINTF("%s(%i):MSCGenerateKeys: %d %d\n",__FILE__,__LINE__, prvKeyNum, pubKeyNum);
    /* 
     * DC: Musclecard has each key number as a unique number but
     * keys should be in sequence - priv 0, pub 1 ... 
     * Here we will set the pubKeyNum to be the same as the 
     * private key number, but public keys will be one + the
     * private key number
     */

    if ( pubKeyNum > AUTHENTIC_MAXMSC_KEY_NUM || 
                            prvKeyNum > AUTHENTIC_MAXMSC_KEY_NUM )
        return MSC_INVALID_PARAMETER;

    if ( pParams->algoType != MSC_GEN_ALG_RSA_CRT ) 
        return MSC_INVALID_PARAMETER;
  
    /* 
     * Select key container directory.
     */
    SELECT_MF;

    MSC_RETURN_IF_ERROR( auSelectFile( pConnection, AUTHENTIC_DIR_ID,
            SELECT_MODE_DF|SELECT_MODE_WITH_ANSWER,
            response, &responseLen));
    
    MSC_RETURN_IF_ERROR( auSelectFile( pConnection, AUTHENTIC_KEYDIR_ID,
            SELECT_MODE_DF|SELECT_MODE_WITH_ANSWER,
            response, &responseLen));
   
    /*
     * Create container for the private key.
     * Key length 1024 bits.
     * FileID is equal to  AUTHENTIC_KEY_ID + prvKeyNum.
     * Delete FileID, if exist.
     */
    rv = auCreateFile(pConnection, FILE_TYPE_RSA_PRIVATE, prvKeyID, 
            0x0400, privateKeyACL);
    if (rv == MSC_OBJECT_EXISTS)  { 
        MSC_RETURN_IF_ERROR( auDeleteFile ( pConnection, prvKeyID, 
                    FILE_TYPE_RSA_PRIVATE));
        
        MSC_RETURN_IF_ERROR( auCreateFile(pConnection, FILE_TYPE_RSA_PRIVATE, 
                    prvKeyID, 0x0400, privateKeyACL));
    }
     
    /* The same for the public key. */
    rv = auCreateFile(pConnection, FILE_TYPE_RSA_PUBLIC, pubKeyID, 0x0400, 
            publicKeyACL);
    if (rv == MSC_OBJECT_EXISTS)   {
        MSC_RETURN_IF_ERROR( auDeleteFile ( pConnection, pubKeyID, 
                    FILE_TYPE_RSA_PUBLIC));
        
        MSC_RETURN_IF_ERROR( auCreateFile(pConnection, FILE_TYPE_RSA_PUBLIC, 
                    pubKeyID, 0x0400, publicKeyACL));
    }
    
    /* Generate keys */
    MSC_RETURN_IF_ERROR( auGenerateKeyPair( pConnection, prvKeyID, pubKeyID, 0x80));

    /* Update key info record */   
    privACL.readPermission  = MSC_AUT_NONE;
    privACL.writePermission = MSC_AUT_PIN_1;
    privACL.usePermission   = MSC_AUT_PIN_1;
    MSC_RETURN_IF_ERROR( auWriteKeyInfo( pConnection, 
                prvKeyNum, 
                MSC_KEY_RSA_PRIVATE, 
                pubKeyNum, 
                pParams->keySize, 
                &pParams->privateKeyPolicy,
                &privACL));

    pubACL.readPermission  = MSC_AUT_ALL;
    pubACL.writePermission = MSC_AUT_PIN_1;
    pubACL.usePermission   = MSC_AUT_ALL;
    MSC_RETURN_IF_ERROR( auWriteKeyInfo( pConnection, 
                pubKeyNum, 
                MSC_KEY_RSA_PUBLIC, 
                prvKeyNum,
                pParams->keySize, 
                &pParams->privateKeyPolicy,
                &pubACL));

    return rv;
}


MSC_RV PL_MSCImportKey( MSCLPTokenConnection pConnection, MSCUChar8 keyNum,
        MSCLPKeyACL pKeyACL,
        MSCPUChar8 pKeyBlob, MSCULong32 keyBlobSize,
        MSCLPKeyPolicy keyPolicy, MSCPVoid32 pAddParams,
        MSCUChar8 addParamsSize ) 
{
#define SET_BIGNUM(bn,offs) {                                       \
        int size =  pKeyBlob[(offs)]*0x100 + pKeyBlob[(offs)+1];    \
        (bn) = BN_bin2bn(pKeyBlob + (offs) + 2, size, NULL);        \
        DPRINTF("%s(%i): RSA: offs %X; size %X; bn %p\n",           \
                __FILE__,__LINE__,offs,size,(bn));                  \
        }
    
    MSCLong32 rv;
    MSCUChar8 keyType;
    MSCUChar8 keyPartner;
    MSCUShort16 keySize;
    MSCULong32 expSize;
    RSA *rsa;
//    BIGNUM *bn;
    int size;

    DPRINTF("%s(%i): MSCImportKey\n", __FILE__,__LINE__);
    
//TODO Make more clever distinction between RSA and Blob
    if (pAddParams && addParamsSize==2 && 
            (*((MSCPUChar8)pAddParams + 0)==MSC_KEY_RSA_PUBLIC || 
            *((MSCPUChar8)pAddParams + 0)==MSC_KEY_RSA_PRIVATE_CRT) 
            )   {
        DPRINTF("%s(%i): MSCImportKey\n", __FILE__,__LINE__);
        rsa = (RSA *)pKeyBlob;
    }
    else  {
        DPRINTF("%s(%i): MSCImportKey\n", __FILE__,__LINE__);
        rsa = RSA_new();
        keyType   = pKeyBlob[OFFSET_KEYTYPE];
        MemCopyTo16(&keySize, &pKeyBlob[OFFSET_KEYSIZE]);

        if(keyType == MSC_KEY_RSA_PUBLIC)   {
            DPRINTF("%s(%i): MSCImportKey\n", __FILE__,__LINE__);
            SET_BIGNUM(rsa->e, MC_1024P_EXP);
            SET_BIGNUM(rsa->n, MC_1024P_MOD);
        
            //TODO
            keyPartner = keyNum-1;
        }
        else if (keyType == MSC_KEY_RSA_PRIVATE_CRT)  {
            DPRINTF("%s(%i): MSCImportKey\n", __FILE__,__LINE__);
            SET_BIGNUM(rsa->p, MC_1024_OFFSET_P);
            SET_BIGNUM(rsa->q, MC_1024_OFFSET_Q);
            SET_BIGNUM(rsa->iqmp, MC_1024_OFFSET_PQ);
            SET_BIGNUM(rsa->dmp1, MC_1024_OFFSET_DP1);
            SET_BIGNUM(rsa->dmq1, MC_1024_OFFSET_DQ1);
        
            //TODO
            keyPartner = keyNum+1;
        }
        else   {
            DPRINTF("%s(%i): MSCImportKey: Invalid file type %i\n",
                 __FILE__,__LINE__, keyType);
            return MSC_UNSUPPORTED_FEATURE;
        }
    }
  
    DPRINTF("%s(%i): MSCImportKey\n", __FILE__,__LINE__);
    rv = auImportRSAKey(pConnection, keyNum, 
            pKeyACL, rsa, keyPolicy, keyType, keyPartner);

    DPRINTF("%s(%i): MSCImportKey: return %X\n", __FILE__,__LINE__,rv);
    return rv;
}


/*
 * Export RSA key public component.
 */
MSC_RV PL_MSCExportKey( MSCLPTokenConnection pConnection, MSCUChar8 keyNum,
		     MSCPUChar8 pKeyBlob, MSCPULong32 keyBlobSize, 
		     MSCPVoid32 pAddParams, MSCUChar8 addParamsSize ) 
{
    MSCLong32 rv;
    MSCKeyInfo keyInfo;
    MSCUChar8 response[MSC_MAXSIZE_BUFFER];
    MSCUShort16 responseLen;
    MSCUChar8 pExp[MSC_EXP_SIZE];
    MSCUChar8 pMod[MSC_MOD_SIZE];
    char str[MSC_MOD_SIZE*2 + 0x100];
    FILE *ff;
    MSCUShort16 currentVal;

    DPRINTF("%s(%i): MSCExportKey: %i\n",__FILE__,__LINE__,keyNum);

    if ( keyBlobSize==NULL || pKeyBlob==NULL ) {
        DPRINTF("%s(%i): MSCExportKey: Invalide arguments: %p %p\n",__FILE__,__LINE__,
                pKeyBlob,keyBlobSize);
        return MSC_INVALID_PARAMETER;
    }
    
    /* Select key container.  */
    SELECT_MF;
    
    MSC_RETURN_IF_ERROR( auSelectFile( pConnection, 
                AUTHENTIC_DIR_ID,
                SELECT_MODE_DF|SELECT_MODE_WITH_ANSWER,
                response, &responseLen));
    
    MSC_RETURN_IF_ERROR( auSelectFile( pConnection, 
                AUTHENTIC_KEYDIR_ID,
                SELECT_MODE_DF|SELECT_MODE_WITH_ANSWER,
                response, &responseLen));
    
    MSC_RETURN_IF_ERROR( auSelectFile( pConnection, 
                AUTHENTIC_KEY_ID + keyNum,
                SELECT_MODE_EF|SELECT_MODE_WITH_ANSWER,
                response, &responseLen));

    /* Get public key components. */
    MSC_RETURN_IF_ERROR( auReadPublicKey( pConnection, 
                AUTHENTIC_KEY_ID + keyNum,
		        pExp, pMod));

    DPRINTF("%s(%i): MSCExportKey: %i\n",__FILE__,__LINE__,keyNum);

    /* Get key info and check it. */
    rv = auReadKeyInfo(pConnection,keyNum,&keyInfo);
    if(rv!=MSC_SUCCESS)
        return rv;
      
    DPRINTF("%s(%i): MSCExportKey: %i\n",__FILE__,__LINE__,keyNum);
    if ( keyInfo.keyType != MSC_KEY_RSA_PUBLIC ) 
        return MSC_UNSUPPORTED_FEATURE;
    
    /* pkcs11 format */
    pKeyBlob[OFFSET_ENCODING] = MSC_BLOB_ENC_PLAIN;
    pKeyBlob[OFFSET_KEYTYPE]  = MSC_KEY_RSA_PUBLIC;
        
    currentVal = CF_1024_KEYSIZE;
    MemCopy16(&pKeyBlob[OFFSET_KEYSIZE], &currentVal);
        
    currentVal = CF_1024P_MODSIZE;
    MemCopy16(&pKeyBlob[MC_1024P_MOD], &currentVal);

    memcpy(&pKeyBlob[MC_1024P_MOD + 2], pMod,CF_1024P_MODSIZE);

    currentVal = CF_1024P_EXPSIZE;
    MemCopy16(&pKeyBlob[MC_1024P_EXP], &currentVal);
    memcpy(&pKeyBlob[MC_1024P_EXP + 2],
            pExp + MSC_EXP_SIZE - CF_1024P_EXPSIZE, 
            CF_1024P_EXPSIZE);
    *keyBlobSize = MC_1024P_FULLSIZE;
      
    return MSC_SUCCESS;
}



/* 
 * Interface to Encrypt/Decrypt facilities of the x509 applet.
 * Data length to encrypt has to be not more then 51 bytes.
 */
MSC_RV PL_MSCComputeCrypt( MSCLPTokenConnection pConnection,
			MSCLPCryptInit cryptInit, MSCPUChar8 pInputData,
			MSCULong32 inputDataSize, MSCPUChar8 pOutputData,
			MSCPULong32 outputDataSize ) 
{
    MSCLong32 rv;
    MSCTransmitBuffer tBuffer;
    MSCUChar8 cflKeyNum;
    
    MSCUChar8 response[MSC_MAXSIZE_BUFFER];
    MSCUShort16 responseLen;

    DPRINTF("%s(%i): MSCComputeCrypt: %i; key %i\n",__FILE__,__LINE__,
            inputDataSize, cryptInit->keyNum);

    if( cryptInit->cipherDirection == MSC_DIR_ENCRYPT &&  inputDataSize>51)  {
        DPRINTF("%s(%i): MSCComputeCrypt: invalid data size\n",
                __FILE__,__LINE__);
        return  MSC_INVALID_PARAMETER;
    }
    
    if( cryptInit->cipherDirection == MSC_DIR_DECRYPT &&  
            inputDataSize != AUTHENTIC_KEY_SIZE)   {
        DPRINTF("%s(%i): MSCComputeCrypt: invalid data size\n",
                __FILE__,__LINE__);
        return  MSC_INVALID_PARAMETER;
    }

    if ( cryptInit->cipherMode != MSC_MODE_RSA_PAD_PKCS1 ) {
        DPRINTF("%s(%i): MSCComputeCrypt: "
                "cipherMode should be MSC_MODE_RSA_PAD_PKCS1\n",
                __FILE__,__LINE__);
        return  MSC_INVALID_PARAMETER;
    }
        
    
    /* Select key container. */
    SELECT_MF;

    MSC_RETURN_IF_ERROR( auSelectFile( pConnection, 
                AUTHENTIC_DIR_ID,
                SELECT_MODE_DF|SELECT_MODE_WITH_ANSWER,
                response, &responseLen));
    
    MSC_RETURN_IF_ERROR( auSelectFile( pConnection, 
                AUTHENTIC_KEYDIR_ID,
                SELECT_MODE_DF|SELECT_MODE_WITH_ANSWER,
                response, &responseLen));
    
    MSC_RETURN_IF_ERROR( auSelectFile( pConnection, 
                AUTHENTIC_KEY_ID + cryptInit->keyNum,
                SELECT_MODE_EF|SELECT_MODE_WITH_ANSWER,
                response, &responseLen));
    
    if ( cryptInit->cipherDirection == MSC_DIR_DECRYPT ||
         cryptInit->cipherDirection == MSC_DIR_VERIFY) {

        if(response[2]==0x82 && response[4]==FILE_TYPE_RSA_PUBLIC)   
            MSC_RETURN_IF_ERROR( auDecryptPublic(pConnection, 
                        cryptInit->keyNum, 
                        pInputData, inputDataSize,
                        pOutputData, outputDataSize));
        else   
            MSC_RETURN_IF_ERROR( auDecrypt(pConnection, 
                        cryptInit->keyNum, 
                        pInputData, inputDataSize,
                        pOutputData, outputDataSize));
        
        return rv;
    }
    else if (cryptInit->cipherDirection == MSC_DIR_ENCRYPT ||
             cryptInit->cipherDirection == MSC_DIR_SIGN)   {
        
        if( response[2]==0x82 && response[4]==FILE_TYPE_RSA_PUBLIC)   
             MSC_RETURN_IF_ERROR( auEncryptPublic(pConnection, 
                         cryptInit->keyNum,
                         pInputData, inputDataSize,
                         pOutputData, outputDataSize));
        else   
             MSC_RETURN_IF_ERROR( auEncrypt(pConnection, 
                         cryptInit->keyNum, 
                         pInputData, inputDataSize,
                         pOutputData, outputDataSize));
        
        return rv;
    }
    else   {

        return MSC_INVALID_PARAMETER;
    }
}


MSC_RV PL_MSCExtAuthenticate( MSCLPTokenConnection pConnection, 
        MSCUChar8 keyNum,
        MSCUChar8 cipherMode, MSCUChar8 cipherDirection,
		MSCPUChar8 pData, MSCULong32 dataSize ) {

  DPRINTF("%s(%i): MSCExtAuthenticate\n",__FILE__,__LINE__);
  return MSC_UNSUPPORTED_FEATURE;
}



MSC_RV PL_MSCListKeys( MSCLPTokenConnection pConnection, MSCUChar8 seqOption,
                       MSCLPKeyInfo pKeyInfo )
{
    MSCLong32 rv;
    MSCUChar8 response[MSC_MAXSIZE_BUFFER];
    MSCUShort16 responseLen;
    MSCUShort16 id;
    static int sequenceNumber;
    int valid_name = strcmp(lastSelectedName,"KeyInfo")==0;
    MSCLPKeyInfo ki = NULL;
   
    DPRINTF("%s(%i): MSCListKeys\n",__FILE__,__LINE__);
        
    if (seqOption == MSC_SEQUENCE_NEXT)
        sequenceNumber++;
    
    if (seqOption == MSC_SEQUENCE_RESET)
        sequenceNumber = 1;
    
    if ( getObjectID(pConnection, &id, "KeyInfo",valid_name) != MSC_SUCCESS)
        return MSC_INVALID_PARAMETER;

    if (!valid_name)   
         MSC_RETURN_IF_ERROR( auSelectFile( pConnection, id, 
                SELECT_MODE_EF | SELECT_MODE_WITH_ANSWER,
                &response[0], &responseLen));
    
    MSC_RETURN_IF_ERROR( auReadRecord( pConnection, 
                sequenceNumber,  sizeof(MSCKeyInfo), 
                &response[0],  &responseLen));

#if 0    
    memcpy(pKeyInfo,response,responseLen);
#else
    if (responseLen < sizeof(MSCKeyInfo))
        return MSC_UNSPECIFIED_ERROR;
    
    ki = (MSCLPKeyInfo)response;
    
    pKeyInfo->keyNum =  ki->keyNum;
    pKeyInfo->keyType = ki->keyType;
    pKeyInfo->keyPartner = ki->keyPartner;
    pKeyInfo->keyMapping = ki->keyMapping;
    
    MemCopyTo16(&pKeyInfo->keySize, 
            &ki->keySize);
    
    MemCopyTo16(&pKeyInfo->keyPolicy.cipherMode, 
            &ki->keyPolicy.cipherMode);
    MemCopyTo16(&pKeyInfo->keyPolicy.cipherDirection, 
            &ki->keyPolicy.cipherDirection);

    MemCopyTo16(&pKeyInfo->keyACL.readPermission, 
            &ki->keyACL.readPermission); 
    MemCopyTo16(&pKeyInfo->keyACL.writePermission, 
            &ki->keyACL.writePermission );
    MemCopyTo16(&pKeyInfo->keyACL.usePermission, 
            &ki->keyACL.usePermission);
#endif    

    strcpy(lastSelectedName,"KeyInfo");
    
    return  MSC_SUCCESS;
}


MSC_RV PL_MSCCreatePIN( MSCLPTokenConnection pConnection, 
        MSCUChar8 pinNum,
		MSCUChar8 pinAttempts, MSCPUChar8 pPinCode,
		MSCULong32 pinCodeSize, MSCPUChar8 pUnblockCode,
		MSCUChar8 unblockCodeSize ) 
{
    DPRINTF("%s(%i): MSCCreatePIN\n",__FILE__,__LINE__);
    return MSC_UNSUPPORTED_FEATURE;
}


/* 
 * Check PIN code:
 * Do the verification if PIN or SOPIN had yet been verified,
 * or performs actual verification. 
 * Code length is not more then 15 bytes.
 * 
 * Internally padded with 0xFF untill the final length 15 bytes.
 */
MSC_RV PL_MSCVerifyPIN( MSCLPTokenConnection pConnection, 
        MSCUChar8 pinNumber,
		MSCPUChar8 pPinCode, MSCULong32 pinCodeSize ) 
{
    MSCLong32 rv;
    MSCTransmitBuffer tBuffer;
    MSCPUChar8 apduResponse;
    MSCPUChar8 pBuffer;
    MSCUChar8 pinN;
    int ii;

    DPRINTF("%s(%i): MSCVerifyPIN\n",   __FILE__,__LINE__);
  
    if ( (pPinCode!=NULL && (pinCodeSize > 15 ||  pinCodeSize <0)) ||
         (pinNumber != NUM_SOPIN && pinNumber != NUM_PIN) )   {
        DPRINTF("%s(%i):Invalid Parameter: pinNum %d; codeSize: %d\n",
                __FILE__,__LINE__, pinNumber, pinCodeSize);
        return MSC_INVALID_PARAMETER;
    }
    
    pinN = authentICPins[pinNumber];

    pBuffer = tBuffer.pBuffer; 
    apduResponse = tBuffer.apduResponse;

    /* Check previous verification. */
    pBuffer[OFFSET_CLA]    = CLA_00;
    pBuffer[OFFSET_INS]    = INS_VERIFY_PIN;
    pBuffer[OFFSET_P1]     = 0x00;
    pBuffer[OFFSET_P2]     = pinN;
    pBuffer[OFFSET_LC]     = 0x00;
    tBuffer.bufferSize =  5;
    
    tBuffer.apduResponseSize = MSC_MAXSIZE_BUFFER;

    SCARD_RETURN_IF_ERROR( auSCardSendAPDU(pConnection, &tBuffer));

    if ( tBuffer.apduResponseSize != 2 ) 
        return MSC_UNSPECIFIED_ERROR;
        
    if (tBuffer.apduResponse[0] == AUMSC_SUCCESS_BYTE)
        return MSC_SUCCESS;

    if (pPinCode==NULL)   
        return convertSW(apduResponse);

    /* Perform verification. */
    pBuffer[OFFSET_LC] = 0x0F;
    memcpy(&pBuffer[OFFSET_DATA], pPinCode, pinCodeSize);
    
    for (ii = pinCodeSize;ii<0x0F;ii++)
        pBuffer[OFFSET_DATA+ii] = 0xFF;
    
    tBuffer.bufferSize = 0x14;

    tBuffer.apduResponseSize = MSC_MAXSIZE_BUFFER;

    SCARD_RETURN_IF_ERROR( auSCardSendAPDU(pConnection, &tBuffer));
  
    AUTENTHIC_COMMON_COMMAND_RETURN;
}


/* Change PIN or SOPIN codes. */
MSC_RV PL_MSCChangePIN( MSCLPTokenConnection pConnection, 
        MSCUChar8 pinNumber,
		MSCPUChar8 pOldPinCode, MSCUChar8 oldPinCodeSize,
		MSCPUChar8 pNewPinCode, MSCUChar8 newPinCodeSize ) 
{
    MSCLong32 rv;
    MSCTransmitBuffer tBuffer;
    MSCPUChar8 apduResponse;
    MSCPUChar8 pBuffer;
    MSCUChar8 pinN;
    int ii;
    
    DPRINTF("%s(%i): MSCChangePIN\n", __FILE__,__LINE__);
    
    if ( oldPinCodeSize > 15 || newPinCodeSize > 15 ||
         (pinNumber != NUM_SOPIN && pinNumber != NUM_PIN)    )   {
        DPRINTF("%s(%i):Invalid Parameter \n", __FILE__,__LINE__);
        return MSC_INVALID_PARAMETER;
    }

    pinN = authentICPins[pinNumber];

    pBuffer = tBuffer.pBuffer; 
    apduResponse = tBuffer.apduResponse;

    pBuffer[OFFSET_CLA] = CLA_00;
    pBuffer[OFFSET_INS] = INS_REFERENCE;
    pBuffer[OFFSET_P1]  = 0x00;
    pBuffer[OFFSET_P2]  = pinN;
    pBuffer[OFFSET_LC]  = 0x1E;
    tBuffer.bufferSize  = 0x23;
    tBuffer.apduResponseSize = MSC_MAXSIZE_BUFFER;
    
    memcpy(&pBuffer[OFFSET_DATA], pOldPinCode, oldPinCodeSize);
    for (ii = oldPinCodeSize;ii<0x0F;ii++)
        pBuffer[OFFSET_DATA+ii] = 0xFF;

    memcpy(&pBuffer[OFFSET_DATA + 0x0F], pNewPinCode, newPinCodeSize);
    for (ii = newPinCodeSize; ii<0x0F; ii++)
        pBuffer[OFFSET_DATA + 0x0F + ii] = 0xFF;

    SCARD_RETURN_IF_ERROR( auSCardSendAPDU(pConnection, &tBuffer));

    if ( tBuffer.apduResponseSize != 2 ) 
        return MSC_UNSPECIFIED_ERROR;
        
    if (tBuffer.apduResponse[0] == AUMSC_SUCCESS_BYTE)
        return MSC_SUCCESS;

    AUTENTHIC_COMMON_COMMAND_RETURN;
}


/* Reset retry counter of the PIN. */
MSC_RV PL_MSCUnblockPIN( MSCLPTokenConnection pConnection, MSCUChar8 pinNumber,
		MSCPUChar8 pUnblockCode, MSCULong32 unblockCodeSize)
{
    return auUnblockPIN(pConnection, pinNumber, 
            pUnblockCode, unblockCodeSize,
            NULL, 0);
}


/* There are two PINs: SOPIN and PIN.*/
MSC_RV PL_MSCListPINs( MSCLPTokenConnection pConnection, 
		    MSCPUShort16 pPinBitMask ) 
{
  
  DPRINTF("%s(%i): MSCListPINs\n",__FILE__,__LINE__);
  *pPinBitMask = 0x00000003;
  return MSC_SUCCESS;
}



/* 
 * Create file and write its information to ObjectInfo.
 */
MSC_RV PL_MSCCreateObject( MSCLPTokenConnection pConnection, 
        MSCString objectID,
        MSCULong32 objectSize, MSCLPObjectACL pObjectACL ) 
{
    MSCUShort16 id;
    MSCLong32 rv;
    MSCObjectInfo objectInfo;
    int valid_name;

    MSCUChar8 acl[8];    
        
    DPRINTF("%s(%i): MSCCreateObject %s\n",__FILE__,__LINE__, objectID);

    /* Imlicitely select the object: do not use cacheObjectName */
    valid_name = 0;
    
    if ( getObjectID(pConnection, &id, objectID,valid_name) != MSC_SUCCESS)
        return MSC_INVALID_PARAMETER;

    getOberthurACL(acl, (MSCPUChar8) pObjectACL);

    DPRINTF("%s(%i): MSCCreateObject %s\n",__FILE__,__LINE__, objectID);
    MSC_RETURN_IF_ERROR( auCreateFile(pConnection, 
                FILE_TYPE_EF_TRANSPARENT,
                id, objectSize, acl));
    
    DPRINTF("%s(%i): MSCCreateObject %s\n",__FILE__,__LINE__, objectID);
    /* Write Object info   */ 
    auDeleteObjectInfo( pConnection, objectID);
    
    DPRINTF("%s(%i): MSCCreateObject %s\n",__FILE__,__LINE__, objectID);
    strncpy(objectInfo.objectID,objectID,MSC_MAXSIZE_OBJID);
    objectInfo.objectSize = objectSize;
    objectInfo.objectACL.readPermission = pObjectACL->readPermission;
    objectInfo.objectACL.writePermission = pObjectACL->writePermission;
    objectInfo.objectACL.deletePermission = pObjectACL->deletePermission;
    rv = auAppendObjectInfo( pConnection, &objectInfo);

    DPRINTF("%s(%i): MSCCreateObject %s\n",__FILE__,__LINE__, objectID);
    /* Clear cacheObjectName. */
    strcpy(lastSelectedName,"");

    return rv;
}


/* Delete file and corresponding record in the ObjectInfo. */
MSC_RV PL_MSCDeleteObject( MSCLPTokenConnection pConnection, 
		MSCString objectID, MSCUChar8 zeroFlag ) 
{
    MSCUShort16 id;
    MSCLong32 rv;
    /* Try to use cacheObjectName.*/
    int valid_name = strcmp(lastSelectedName,objectID)==0;
        
    DPRINTF("%s(%i): MSCDeleteObject\n",__FILE__,__LINE__);
    if (getObjectID(pConnection, &id, objectID, valid_name) != MSC_SUCCESS)
        return MSC_INVALID_PARAMETER;

    MSC_RETURN_IF_ERROR( auDeleteFile(pConnection, 
                id, 
                FILE_TYPE_EF_TRANSPARENT));
    
    auDeleteObjectInfo( pConnection, objectID);

    return rv;
}


/* Update binary file. */
MSC_RV PL_MSCWriteObject( MSCLPTokenConnection pConnection, 
        MSCString objectID,
		MSCULong32 offset, MSCPUChar8 pInputData, MSCUChar8 dataSize)
{
    MSCLong32 rv, rv_msc;	
    MSCUShort16 id;
    MSCUChar8  response[MAX_BUFFER_SIZE];
    MSCUShort16 responseLen;
    int obj, jj;
    int valid_name = strcmp(lastSelectedName,objectID)==0;

    DPRINTF("%s(%i): MSCWriteObject %s: valid_name %i; offs 0x%lX, size 0x%lX\n",
            __FILE__,__LINE__,objectID,valid_name,offset,dataSize);
    if ( getObjectID(pConnection,&id, objectID, valid_name) != MSC_SUCCESS)
        return MSC_INVALID_PARAMETER;
       
    if (!valid_name)   {
        MSC_RETURN_IF_ERROR( auSelectFile( pConnection, id, 
                    SELECT_MODE_EF | SELECT_MODE_WITH_ANSWER,
                    &response[0], &responseLen));
    
        if ( *(response+8) != FILE_TYPE_EF_TRANSPARENT )   
            return MSC_INVALID_PARAMETER;
    }
    
    strcpy(lastSelectedName,objectID);
    
    MSC_RETURN_IF_ERROR( auUpdateBinaryEF( pConnection, offset, 
                pInputData, dataSize ));

    return rv;
}

MSC_RV PL_MSCReadObject( MSCLPTokenConnection pConnection, MSCString objectID, 
		      MSCULong32 offset, MSCPUChar8 pOutputData, MSCUChar8 dataSize)
{
    MSCLong32 rv;	
    MSCUShort16 id;
    MSCUChar8  response[MAX_BUFFER_SIZE];
    MSCUShort16 responseLen;
    int obj, jj;
    int valid_name = strcmp(lastSelectedName,objectID)==0;
    
    DPRINTF("%s(%i): MSCReadObject: id %s; offs Ox%X; size 0x%X\n",
            __FILE__,__LINE__,objectID, offset, dataSize);
    if ( getObjectID(pConnection, &id, objectID, valid_name) != MSC_SUCCESS)
        return MSC_INVALID_PARAMETER;
       
    MSC_RETURN_IF_ERROR( auSelectFile( pConnection, id, 
                SELECT_MODE_EF | SELECT_MODE_WITH_ANSWER,
                &response[0], &responseLen));
    if (*(response+8) != FILE_TYPE_EF_TRANSPARENT)
         return MSC_INVALID_PARAMETER;

    MSC_RETURN_IF_ERROR( auReadBinary( pConnection, offset, 
                pOutputData, dataSize ));
    
    strcpy(lastSelectedName,objectID);

    return rv;
}


MSC_RV PL_MSCReadAllocateObject(MSCLPTokenConnection pConnection,
        MSCString objectID, MSCPUChar8 *pOutputData,
        MSCPULong32 dataSize,
        LPRWEventCallback rwCallback, MSCPVoid32 addParams)
{
    DPRINTF("%s(%i): MSCReadAllocateObject\n",__FILE__,__LINE__);
    return MSC_UNSUPPORTED_FEATURE;
}


/* 
 * Get next object info.
 * Records pointer can be initialized by seqOption = MSC_SEQUENCE_RESET.
 */
MSC_RV PL_MSCListObjects( MSCLPTokenConnection pConnection, 
        MSCUChar8 seqOption,
        MSCLPObjectInfo pObjectInfo )
{
    MSCLong32 rv;
    MSCUChar8 response[MSC_MAXSIZE_BUFFER];
    MSCUShort16 responseLen;
    MSCUShort16 id;
    static int sequenceNumber;
    int valid_name = strcmp(lastSelectedName,"ObjectInfo")==0;
   
    DPRINTF("%s(%i): MSCListObjects\n",__FILE__,__LINE__);
        
    if (seqOption == MSC_SEQUENCE_NEXT)
        sequenceNumber++;
    
    if (seqOption == MSC_SEQUENCE_RESET)
        sequenceNumber = 1;
    
    if ( getObjectID(pConnection, &id, "ObjectInfo",valid_name) != MSC_SUCCESS)
        return MSC_INVALID_PARAMETER;

    if (!valid_name)   
        MSC_RETURN_IF_ERROR( auSelectFile( pConnection, id, 
                    SELECT_MODE_EF | SELECT_MODE_WITH_ANSWER,
                    &response[0], &responseLen));

    MSC_RETURN_IF_ERROR( auReadRecord( pConnection, 
            sequenceNumber, OBJECT_INFO_SIZE, 
            &response[0],  &responseLen));

    memcpy(pObjectInfo,response,responseLen);

    strcpy(lastSelectedName,"ObjectInfo");
    
    return  MSC_SUCCESS;
}



/* 
 * There is no native logout authentIC applet command.
 * Try to choose some dummy application.
 */
MSC_RV PL_MSCLogoutAll( MSCLPTokenConnection pConnection ) 
{
#if 0    
    struct FileInfo *files;
    struct FileInfo *tree[10];
    int ptr = 0;
    long size = 0;
    
    auListFilesAllocate( pConnection, "3F00", &files);

goHere:
    while(files->prev != files)   {
        DPRINTF("ID %s/%X; type %X, size %X\n",
                 files->path, files->id, files->type, files->size);
        
        size += files->size;
        
        if (files->list != NULL)   {
            tree[ptr++] = files->prev;
            files = files->list;
            continue;
        }
         
        files = files->prev;
    }
    
    if (ptr>0)   {
        files = tree[--ptr];
        goto goHere;
    }
    DPRINTF("%s(%i): MSCLogoutAll: total size %li\n",__FILE__,__LINE__, size);
#else    
    DPRINTF("%s(%i): MSCLogoutAll: not supported feature\n",__FILE__,__LINE__);
#endif
    return MSC_UNSUPPORTED_FEATURE;
}


/*
 * Generate random number.
 */
MSC_RV PL_MSCGetChallenge( MSCLPTokenConnection pConnection, MSCPUChar8 pSeed,
			MSCUShort16 seedSize, MSCPUChar8 pRandomData,
			MSCUShort16 randomDataSize ) 
{

    MSCLong32 rv;
    MSCTransmitBuffer tBuffer;
    MSCPUChar8 apduResponse;
    MSCPUChar8 pBuffer;

    DPRINTF("%s(%i): MSCGetChallenge\n",__FILE__,__LINE__);

    if ( pRandomData == 0  && randomDataSize < 8 ) 
        return MSC_INVALID_PARAMETER;

    pBuffer = tBuffer.pBuffer; 
    apduResponse = tBuffer.apduResponse;

    tBuffer.pBuffer[OFFSET_CLA] = CLA_00;
    tBuffer.pBuffer[OFFSET_INS] = INS_GET_CHALLENGE;
    tBuffer.pBuffer[OFFSET_P1]  = 0x00;
    tBuffer.pBuffer[OFFSET_P2]  = 0x00;
    tBuffer.pBuffer[OFFSET_LC]  = randomDataSize;

    tBuffer.bufferSize = 5;

  /* Set up the APDU exchange */
    tBuffer.apduResponseSize = MSC_MAXSIZE_BUFFER;
    SCARD_RETURN_IF_ERROR( auSCardSendAPDU(pConnection, &tBuffer));
  
    if ( tBuffer.apduResponseSize <= 2 )
        return MSC_UNSPECIFIED_ERROR;
    
    if(tBuffer.apduResponse[tBuffer.apduResponseSize-2] == AUMSC_SUCCESS_BYTE)  {
        memcpy(pRandomData, tBuffer.apduResponse, randomDataSize);
        return MSC_SUCCESS;
    }
    else
        return convertSW(tBuffer.apduResponse + tBuffer.apduResponseSize-2);
}

/********** end of musclecard-framework functions ******************************/



/*
 * Clean smart card authentIC framework.
 */
MSC_RV auCleanFramework( MSCLPTokenConnection pConnection) 
{
    MSCLong32 rv;
    MSCUChar8 response[MSC_MAXSIZE_BUFFER];
    MSCUShort16 responseLen;
    MSCUShort16 fileID;
    
    DPRINTF("%s(%i): MSCCleanFramework\n",__FILE__,__LINE__);
    
    SELECT_AUTHENTIC_DIR;
    
    rv = auSelectFile( pConnection, AUTHENTIC_KEYDIR_ID,
            SELECT_MODE_DF|SELECT_MODE_WITH_ANSWER,
            response, &responseLen);
    if ( rv == MSC_SUCCESS )    {
    
        rv = auListFiles(pConnection, MSC_SEQUENCE_RESET, &fileID);
        for( ;rv == MSC_SUCCESS && fileID!=0xFFFF; ) {
            MSC_RETURN_IF_ERROR( auDeleteFile ( pConnection, fileID, 
                         FILE_TYPE_EF_TRANSPARENT));
            
            rv = auListFiles(pConnection, MSC_SEQUENCE_NEXT, &fileID);
        }
        
        SELECT_AUTHENTIC_DIR;
        MSC_RETURN_IF_ERROR( auDeleteFile ( pConnection, 
                    AUTHENTIC_KEYDIR_ID, FILE_TYPE_DF));
    }
  
    SELECT_AUTHENTIC_DIR; 
    rv = auSelectFile( pConnection, AUTHENTIC_OBJECTDIR_ID,
            SELECT_MODE_DF|SELECT_MODE_WITH_ANSWER,
            response, &responseLen);
    if ( rv == MSC_SUCCESS )   {
    
        rv = auListFiles(pConnection, MSC_SEQUENCE_RESET, &fileID);
        for( ;rv == MSC_SUCCESS && fileID!=0xFFFF; ) {
            MSC_RETURN_IF_ERROR( auDeleteFile ( pConnection, fileID, 
                         FILE_TYPE_EF_TRANSPARENT));

            rv = auListFiles(pConnection, MSC_SEQUENCE_NEXT, &fileID);
        }

        SELECT_AUTHENTIC_DIR;
        MSC_RETURN_IF_ERROR( auDeleteFile ( pConnection, 
                     AUTHENTIC_OBJECTDIR_ID, FILE_TYPE_DF));
    }

    
    SELECT_AUTHENTIC_DIR;
    rv = auListFiles(pConnection, MSC_SEQUENCE_RESET, &fileID);
    for( ;rv == MSC_SUCCESS && fileID!=0xFFFF; ) {
        MSC_RETURN_IF_ERROR( auDeleteFile ( pConnection, fileID, 
                     FILE_TYPE_EF_TRANSPARENT));
        
        rv = auListFiles(pConnection, MSC_SEQUENCE_NEXT, &fileID);
    }
    
    SELECT_MF;
    MSC_RETURN_IF_ERROR( auDeleteFile ( pConnection, 
                AUTHENTIC_DIR_ID, FILE_TYPE_DF));

    return MSC_SUCCESS;
}


/* 
 * Select EF or DF file.
 * TODO Introduce implicit selection.
 * TODO Review file selection without answer.
 */
MSC_RV auSelectFile(MSCLPTokenConnection pConnection,  
        MSCULong32 fileID, MSCUChar8 flag, 
        MSCPUChar8 apduResponse, MSCPUShort16 lenResponse) 
{
    MSCLong32 rv;
    MSCTransmitBuffer tBuffer;
    
    DPRINTF("%s(%i): MSCSelectFile\n",__FILE__,__LINE__);

    tBuffer.pBuffer[OFFSET_CLA] = CLA_00;
    tBuffer.pBuffer[OFFSET_INS] = INS_SELECT_FILE;
    tBuffer.pBuffer[OFFSET_P1]  = flag&0x03;
    tBuffer.pBuffer[OFFSET_P2]  = flag&0x0C;
    if ((flag&0x03) != 0x03)   {
        tBuffer.pBuffer[OFFSET_LC]  = 0x02;
        tBuffer.pBuffer[OFFSET_DATA+0] = (fileID>>8)&0xFF;
        tBuffer.pBuffer[OFFSET_DATA+1] = fileID&0xFF;
        if (flag&0x0C)   
            tBuffer.pBuffer[OFFSET_DATA+2] = 0x00; 
        else
            tBuffer.pBuffer[OFFSET_DATA+2] = 0x18;
        tBuffer.bufferSize = 8;
    }
    else   {
        if (flag&0x0C)
            tBuffer.pBuffer[OFFSET_LC]  = 0x00;
        else
             tBuffer.pBuffer[OFFSET_LC]  = 0x18;
        tBuffer.bufferSize = 5;
    }
    
    
    tBuffer.apduResponseSize = MSC_MAXSIZE_BUFFER;
   
    SCARD_RETURN_IF_ERROR( auSCardExchangeAPDU(pConnection, &tBuffer));
    
    *lenResponse = tBuffer.apduResponseSize;

    strcpy(lastSelectedName,"");
   
    if(tBuffer.apduResponseSize < 2)   
        return MSC_UNSPECIFIED_ERROR;

    if ( apduResponse != NULL)   {
        int ii;
        for (ii=0; ii<tBuffer.apduResponseSize;ii++)   {
            apduResponse[ii] = tBuffer.apduResponse[ii];
        }
    } 

    if(tBuffer.apduResponse[tBuffer.apduResponseSize-2] == AUMSC_SUCCESS_BYTE)
        return MSC_SUCCESS;
    else 
        return convertSW(tBuffer.apduResponse + tBuffer.apduResponseSize-2 );
}


/* Delete specified file. */
MSC_RV auDeleteFile( MSCLPTokenConnection pConnection,
             MSCUShort16 fileID, MSCUChar8 fileType)
{
    MSCLong32 rv;
    MSCTransmitBuffer tBuffer;

    DPRINTF("%s(%i): MSCDeleteFile: id %X\n",__FILE__,__LINE__, fileID);

    tBuffer.pBuffer[OFFSET_CLA]    = CLA_00;
    tBuffer.pBuffer[OFFSET_INS]    = INS_DELETE_FILE;
    if (fileType == FILE_TYPE_DF)
        tBuffer.pBuffer[OFFSET_P1]     = 0x01;
    else 
        tBuffer.pBuffer[OFFSET_P1]     = 0x02;
        
    tBuffer.pBuffer[OFFSET_P2]     = 0x00;
    tBuffer.pBuffer[OFFSET_LC]     = 0x02;
    tBuffer.pBuffer[OFFSET_DATA + 0] = fileID >> 8;
    tBuffer.pBuffer[OFFSET_DATA + 1] = fileID & 0xFF;

    tBuffer.bufferSize = 7;
    tBuffer.apduResponseSize = MSC_MAXSIZE_BUFFER;
    
    /* Clear cacheObjectName. */
    strcpy(lastSelectedName,""); 
    
    SCARD_RETURN_IF_ERROR( auSCardSendAPDU(pConnection, &tBuffer));
    
    AUTENTHIC_COMMON_COMMAND_RETURN;
}


/* 
 * Generate RSA key pair.
 */
MSC_RV auGenerateKeyPair( MSCLPTokenConnection pConnection, 
        MSCUShort16 prvKey, MSCUShort16 pubKey, MSCUChar8 keySize ) 
{
    MSCLong32 rv;
    MSCTransmitBuffer tBuffer;

    DPRINTF("%s(%i): auGenerateKeyPair\n",__FILE__,__LINE__);
    tBuffer.pBuffer[OFFSET_CLA]     = CLA_00;
    tBuffer.pBuffer[OFFSET_INS]     = INS_GENERATE_KEY;
    tBuffer.pBuffer[OFFSET_P1]      = 0x00;
    tBuffer.pBuffer[OFFSET_P2]      = 0x00;
    tBuffer.pBuffer[OFFSET_LC]      = 0x04;

    tBuffer.pBuffer[OFFSET_DATA+0]  = pubKey >> 8;
    tBuffer.pBuffer[OFFSET_DATA+1]  = pubKey & 0xFF;
    tBuffer.pBuffer[OFFSET_DATA+2]  = prvKey >> 8;
    tBuffer.pBuffer[OFFSET_DATA+3]  = prvKey & 0xFF;
    
    tBuffer.pBuffer[OFFSET_DATA+4]  = keySize;
    tBuffer.bufferSize              = 0x0A;
    
    tBuffer.apduResponseSize = MSC_MAXSIZE_BUFFER;

    SCARD_RETURN_IF_ERROR( auSCardExchangeAPDU(pConnection, &tBuffer));

    AUTENTHIC_COMMON_COMMAND_RETURN;
}


/* Interface to ecrypt smart card facility. */
MSC_RV auEncrypt( MSCLPTokenConnection pConnection, MSCUChar8 keyNum,
			MSCPUChar8 pInputData, MSCULong32 inputDataSize, 
            MSCPUChar8 pOutputData, MSCPULong32 outputDataSize) 
{
    MSCLong32 rv;
    MSCTransmitBuffer tBuffer;
    MSCUChar8 response[MSC_MAXSIZE_BUFFER];
    MSCUShort16 responseLen;

    DPRINTF("MSCEncrypt: keyNum %i\n",keyNum);
    /* MSE-Set: set parameters of the Digital Signature Template.*/
    tBuffer.pBuffer[OFFSET_CLA]  = CLA_00;
    tBuffer.pBuffer[OFFSET_INS]  = INS_SIGNATURE_MSE;
    tBuffer.pBuffer[OFFSET_P1]   = 0x41;
    tBuffer.pBuffer[OFFSET_P2]   = 0xB6;
    tBuffer.pBuffer[OFFSET_LC]   = 0x07;
    
    tBuffer.pBuffer[OFFSET_DATA+0] = 0x80;
    tBuffer.pBuffer[OFFSET_DATA+1] = 0x01;
    tBuffer.pBuffer[OFFSET_DATA+2] = 0x11;
    tBuffer.pBuffer[OFFSET_DATA+3] = 0x81;
    tBuffer.pBuffer[OFFSET_DATA+4] = 0x02;
    tBuffer.pBuffer[OFFSET_DATA+5] = AUTHENTIC_KEY_ID >> 8;
    tBuffer.pBuffer[OFFSET_DATA+6] = keyNum;
    
    tBuffer.bufferSize  = 0x0C;
    tBuffer.apduResponseSize = MSC_MAXSIZE_BUFFER;
    
    SCARD_RETURN_IF_ERROR( auSCardSendAPDU(pConnection, &tBuffer));

    if (tBuffer.apduResponseSize != 2)    
        return MSC_UNSPECIFIED_ERROR;
    
    if (tBuffer.apduResponse[0] != AUMSC_SUCCESS_BYTE)  
        return convertSW(tBuffer.apduResponse);

    /* PSO-Compute Digital Signature: Command message. */
    tBuffer.pBuffer[OFFSET_CLA]  = CLA_00;
    tBuffer.pBuffer[OFFSET_INS]  = INS_SIGNATURE_PSO;
    tBuffer.pBuffer[OFFSET_P1]   = 0x9E;
    tBuffer.pBuffer[OFFSET_P2]   = 0x9A;
    tBuffer.pBuffer[OFFSET_LC]   = inputDataSize;
    memcpy(tBuffer.pBuffer + OFFSET_DATA, pInputData, inputDataSize);
    tBuffer.pBuffer[OFFSET_DATA + inputDataSize] = AUTHENTIC_KEY_SIZE;

    tBuffer.bufferSize  = inputDataSize + 6;
    tBuffer.apduResponseSize = MSC_MAXSIZE_BUFFER;
    
    SCARD_RETURN_IF_ERROR( auSCardExchangeAPDU(pConnection, &tBuffer));

    if (tBuffer.apduResponseSize < 2 || 
                tBuffer.apduResponseSize != AUTHENTIC_KEY_SIZE+2)
        return MSC_UNSPECIFIED_ERROR;
    
    if (tBuffer.apduResponse[tBuffer.apduResponseSize - 2] != AUMSC_SUCCESS_BYTE)
        return convertSW(tBuffer.apduResponse + tBuffer.apduResponseSize - 2);
   
     memcpy(pOutputData, tBuffer.apduResponse, AUTHENTIC_KEY_SIZE);
     *outputDataSize  = AUTHENTIC_KEY_SIZE;
     return  MSC_SUCCESS;
}



MSC_RV auEncryptPublic( MSCLPTokenConnection pConnection, MSCUChar8 keyNum,
			MSCPUChar8 pInputData, MSCULong32 inputDataSize, 
            MSCPUChar8 pOutputData, MSCPULong32 outputDataSize) 
{
    MSCLong32 rv;
    MSCKeyInfo keyInfo;
    RSA *rsa;
                    
    DPRINTF("%s(%i): MSCEncryptPublic: keyNum %i\n",
            __FILE__,__LINE__,keyNum);
    
    MSC_RETURN_IF_ERROR( auGetPublicRSAKey(pConnection, keyNum, &rsa));
    
    rv = RSA_public_encrypt(inputDataSize, pInputData,
                            pOutputData , rsa,  RSA_PKCS1_PADDING);
    if (rv == -1)   {
        DPRINTF("%s(%i): MSCEncryptPublic: cannot encrypt.\n", 
                __FILE__,__LINE__);
        return MSC_UNSPECIFIED_ERROR;
    }
            
    RSA_free(rsa);    
    
    *outputDataSize = AUTHENTIC_KEY_SIZE;
    return MSC_SUCCESS;
}


/* Interface to ecrypt smart card facility. */
MSC_RV auDecrypt( MSCLPTokenConnection pConnection, MSCUChar8 keyNum,
			MSCPUChar8 pInputData, MSCULong32 inputDataSize, 
            MSCPUChar8 pOutputData, MSCPULong32 outputDataSize) 
{
    MSCLong32 rv;
    MSCTransmitBuffer tBuffer;
    MSCUChar8 response[MSC_MAXSIZE_BUFFER];
    MSCUShort16 responseLen;

    DPRINTF("%i: keyNum %i\n",__LINE__,keyNum);
    /* MSE-Set: set confidentiality Template*/
    tBuffer.pBuffer[OFFSET_CLA]  = CLA_00;
    tBuffer.pBuffer[OFFSET_INS]  = INS_SIGNATURE_MSE;
    tBuffer.pBuffer[OFFSET_P1]   = 0xC1;
    tBuffer.pBuffer[OFFSET_P2]   = 0xB8;
    tBuffer.pBuffer[OFFSET_LC]   = 0x07;
    
    tBuffer.pBuffer[OFFSET_DATA+0] = 0x80;
    tBuffer.pBuffer[OFFSET_DATA+1] = 0x01;
    tBuffer.pBuffer[OFFSET_DATA+2] = 0x02;
    tBuffer.pBuffer[OFFSET_DATA+3] = 0x81;
    tBuffer.pBuffer[OFFSET_DATA+4] = 0x02;
    tBuffer.pBuffer[OFFSET_DATA+5] = AUTHENTIC_KEY_ID >> 8;
    tBuffer.pBuffer[OFFSET_DATA+6] = keyNum;
    
    tBuffer.bufferSize  = 0x0C;
    tBuffer.apduResponseSize = MSC_MAXSIZE_BUFFER;
    
    SCARD_RETURN_IF_ERROR( auSCardSendAPDU(pConnection, &tBuffer));
    
    if (tBuffer.apduResponseSize != 2)    
        return MSC_UNSPECIFIED_ERROR;
    
    if (tBuffer.apduResponse[0] != AUMSC_SUCCESS_BYTE)  
        return convertSW(tBuffer.apduResponse);

    /* PSO-Decipher - command message.*/
    tBuffer.pBuffer[OFFSET_CLA]  = CLA_00;
    tBuffer.pBuffer[OFFSET_INS]  = INS_SIGNATURE_PSO;
    tBuffer.pBuffer[OFFSET_P1]   = 0x80;
    tBuffer.pBuffer[OFFSET_P2]   = 0x82;
    tBuffer.pBuffer[OFFSET_LC]   = inputDataSize;
    memcpy(tBuffer.pBuffer + OFFSET_DATA, pInputData, inputDataSize);
    tBuffer.pBuffer[OFFSET_DATA + inputDataSize] = AUTHENTIC_KEY_SIZE;

    tBuffer.bufferSize  = inputDataSize + 6;
    tBuffer.apduResponseSize = MSC_MAXSIZE_BUFFER;
    
//    SCARD_RETURN_IF_ERROR( auSCardSendAPDU(pConnection, &tBuffer));
    SCARD_RETURN_IF_ERROR( auSCardExchangeAPDU(pConnection, &tBuffer));

    if (tBuffer.apduResponseSize < 2)
        return MSC_UNSPECIFIED_ERROR;
    
    if (tBuffer.apduResponse[tBuffer.apduResponseSize - 2] != AUMSC_SUCCESS_BYTE)
        return convertSW(tBuffer.apduResponse + tBuffer.apduResponseSize - 2);

    memcpy(pOutputData, tBuffer.apduResponse, tBuffer.apduResponseSize - 2);
    *outputDataSize  = tBuffer.apduResponseSize - 2;
    return  MSC_SUCCESS;
}


MSC_RV auDecryptPublic( MSCLPTokenConnection pConnection, MSCUChar8 keyNum,
			MSCPUChar8 pInputData, MSCULong32 inputDataSize, 
            MSCPUChar8 pOutputData, MSCPULong32 outputDataSize) 
{
    MSCLong32 rv;
    MSCKeyInfo keyInfo;
    RSA *rsa;
                    
    DPRINTF("MSCDecryptPublic: keyNum %i\n",keyNum);
    
    rv = auGetPublicRSAKey(pConnection, keyNum, &rsa);
    if ( rv != MSC_SUCCESS)
        return rv;

    rv = RSA_public_decrypt(inputDataSize, pInputData,
                            pOutputData , rsa,  RSA_PKCS1_PADDING);
    if (rv == -1)   {
        DPRINTF("MSCDecryptPublic cannot decrypt with %i : %X\n", keyNum,rv);
        return MSC_UNSPECIFIED_ERROR;
    }
            
    RSA_free(rsa);    
    
    *outputDataSize = AUTHENTIC_KEY_SIZE;
    return MSC_SUCCESS;
}



MSC_RV auGetPublicRSAKey(MSCLPTokenConnection pConnection, 
        MSCUChar8 keyNum, RSA **rsaOutput)
{
    MSCLong32 rv;
    MSCKeyInfo keyInfo; 
    int mod_size, exp_size, key_size, ii;
    char str[0x200];
    #define MC_1024P_MOD            4
    #define MC_1024P_EXP          134
    #define KEY_BLOB_SIZE 0x200
    MSCUChar8 pKeyBlob[KEY_BLOB_SIZE];
    MSCULong32 keyBlobSize = KEY_BLOB_SIZE;
    RSA *rsa;
                
    MSC_RETURN_IF_ERROR( auReadKeyInfo(pConnection, keyNum, &keyInfo));
    
    MSC_RETURN_IF_ERROR( PL_MSCExportKey( pConnection, keyNum, pKeyBlob, 
                &keyBlobSize, NULL, 0));
    
    key_size = *(pKeyBlob+2) * 0x100 + *(pKeyBlob+3);
    mod_size = *(pKeyBlob + MC_1024P_MOD) * 0x100 +
                         *(pKeyBlob + MC_1024P_MOD +1);
    exp_size = *(pKeyBlob + MC_1024P_EXP) * 0x100 +
                         *(pKeyBlob + MC_1024P_EXP +1);
    
    rsa = (RSA *)RSA_new();
    rsa->n = BN_new();
    rsa->e = BN_new();
    
    for(ii=0;ii<exp_size;ii++)
        sprintf(str+ii*2,"%02X",*(pKeyBlob + MC_1024P_EXP + 2 + ii));
    BN_hex2bn(&rsa->e, str);
               
    for(ii=0;ii<mod_size;ii++)
        sprintf(str+ii*2,"%02X",*(pKeyBlob + MC_1024P_MOD + 2 + ii));
    BN_hex2bn(&rsa->n, str);

    *rsaOutput = rsa;
    return MSC_SUCCESS;
}



/* Update key info record. */
MSC_RV auWriteKeyInfo( MSCLPTokenConnection pConnection, 
        MSCUChar8 keyNum,  MSCUChar8 keyType, MSCUChar8 keyPartner,
        MSCUShort16 keySize, 
        MSCLPKeyPolicy keyPolicy, 
        MSCLPKeyACL keyACL) 
{
    MSCLong32 rv;
    MSCKeyInfo keyInfo;

    keyInfo.keyNum = keyNum;
    keyInfo.keyType = keyType;
    keyInfo.keyPartner = keyPartner;
    keyInfo.keyMapping = keyNum;

    MemCopy16(&keyInfo.keySize, &keySize);
    
    MemCopy16(&keyInfo.keyPolicy.cipherMode, &keyPolicy->cipherMode);
    MemCopy16(&keyInfo.keyPolicy.cipherDirection, &keyPolicy->cipherDirection);
  
    MemCopy16(&keyInfo.keyACL.readPermission, &keyACL->readPermission);
    MemCopy16(&keyInfo.keyACL.writePermission, &keyACL->writePermission);
    MemCopy16(&keyInfo.keyACL.usePermission, &keyACL->usePermission);
    
    auDeleteKeyInfo( pConnection, keyNum);
    MSC_RETURN_IF_ERROR( auAppendRecord( pConnection,
		       (MSCPUChar8) &keyInfo, sizeof(MSCKeyInfo)));

    /* Clear cacheObjectName. */
    strcpy(lastSelectedName,"");
                                  
    return rv;
}


/* Read record from KeyInfo file. */
MSC_RV auReadKeyInfo( MSCLPTokenConnection pConnection, MSCUChar8 keyNum,  
        MSCLPKeyInfo pKeyInfo)
{
    MSCLong32 rv;
    
    DPRINTF("%s(%i): MSCReadKeyInfo: %i\n",__FILE__,__LINE__,keyNum);
    /* Get key info and check it. */
    pKeyInfo->keyNum = 0xFF;
    rv = PL_MSCListKeys(pConnection, MSC_SEQUENCE_RESET, pKeyInfo);

    while (rv != MSC_SEQUENCE_END ) {
        if ( pKeyInfo->keyNum == keyNum ) 
                break;
        rv = PL_MSCListKeys(pConnection, MSC_SEQUENCE_NEXT, pKeyInfo);
    }

    return rv;
}


/* Delete record from KeyInfo file. */
MSC_RV auDeleteKeyInfo( MSCLPTokenConnection pConnection, MSCUChar8 keyNum)
{
    MSCLong32 rv;
    MSCKeyInfo keyInfo;
    int index;
   
    MSC_RETURN_IF_ERROR(  PL_MSCListKeys( pConnection, MSC_SEQUENCE_RESET, &keyInfo));

    for (index = 1; rv==MSC_SUCCESS; index++ )   {
        if ( keyNum == keyInfo.keyNum)   {
            rv = auDeleteRecord( pConnection, index);
            break;
        }
        
        rv = PL_MSCListKeys( pConnection, MSC_SEQUENCE_NEXT, &keyInfo );
    }

    return rv;
}



/* Delete record from ObjectInfo file. */
MSC_RV auDeleteObjectInfo( MSCLPTokenConnection pConnection, MSCString objectID)
{
    MSCLong32 rv;
    MSCUChar8 response[MSC_MAXSIZE_BUFFER];
    MSCUShort16 responseLen;
    MSCLPObjectInfo pObjectInfo = (MSCLPObjectInfo) response;
    MSCUShort16 id;
    int index;
   
    MSC_RETURN_IF_ERROR( PL_MSCListObjects( pConnection, MSC_SEQUENCE_RESET, 
                 pObjectInfo));

    for (index = 1; rv==MSC_SUCCESS; index++ )   {
        if (strcmp(pObjectInfo->objectID, objectID)==0)   {
            rv = auDeleteRecord( pConnection, index);
            break;
        }
        
        rv = PL_MSCListObjects( pConnection, MSC_SEQUENCE_NEXT, pObjectInfo );
    }

    return rv;
}


/* Append object info record to ObjectInfo file. */
MSC_RV auAppendObjectInfo( MSCLPTokenConnection pConnection, 
        MSCLPObjectInfo pObjectInfo)
{
    MSCLong32 rv;
    MSCObjectInfo objectInfo;
    
    return auAppendRecord( pConnection,
		       (MSCPUChar8) pObjectInfo, OBJECT_INFO_SIZE);
}



/* 
 * Wrappers for the x509 applet functions.
 */
MSC_RV auUpdateBinaryEF( MSCLPTokenConnection pConnection, 
		       MSCULong32 offset, MSCPUChar8 pInputData, MSCUChar8 dataSize ) 
{
    MSCLong32 rv;
    MSCTransmitBuffer tBuffer;

    DPRINTF("%s(%i): MSCWriteBinaryEF\n",__FILE__,__LINE__);
    
    tBuffer.pBuffer[OFFSET_CLA]    = CLA_00;
    tBuffer.pBuffer[OFFSET_INS]    = INS_UPDATE_BINARY;
    tBuffer.pBuffer[OFFSET_P1]     = (offset & 0x7F00) >> 8;
    tBuffer.pBuffer[OFFSET_P2]     = offset & 0x00FF;
    tBuffer.pBuffer[OFFSET_LC]     = dataSize;
  
    memcpy(tBuffer.pBuffer+OFFSET_DATA, pInputData, dataSize);
    tBuffer.bufferSize = dataSize + 5;

    tBuffer.apduResponseSize = MSC_MAXSIZE_BUFFER;
    
    SCARD_RETURN_IF_ERROR( auSCardSendAPDU(pConnection, &tBuffer));
  
    AUTENTHIC_COMMON_COMMAND_RETURN;
}


MSC_RV auAppendRecord( MSCLPTokenConnection pConnection,
		       MSCPUChar8 pInputData, MSCUChar8 dataSize ) 
{
    MSCLong32 rv;
    MSCTransmitBuffer tBuffer;

    DPRINTF("%s(%i): MSCAppendRecord\n",__FILE__,__LINE__);
    
    tBuffer.pBuffer[OFFSET_CLA]    = CLA_00;
    tBuffer.pBuffer[OFFSET_INS]    = INS_APPEND_RECOD;
    tBuffer.pBuffer[OFFSET_P1]     = 0;
    tBuffer.pBuffer[OFFSET_P2]     = 0;
    tBuffer.pBuffer[OFFSET_LC]     = dataSize;
  
    memcpy(tBuffer.pBuffer+OFFSET_DATA, pInputData, dataSize);
    tBuffer.bufferSize = dataSize + 5;

    tBuffer.apduResponseSize = MSC_MAXSIZE_BUFFER;
    
    SCARD_RETURN_IF_ERROR( auSCardSendAPDU(pConnection, &tBuffer));
  
    AUTENTHIC_COMMON_COMMAND_RETURN;
}


MSC_RV auUpdateRecord( MSCLPTokenConnection pConnection, 
		       MSCUChar8 index, MSCPUChar8 pInputData, MSCUChar8 dataSize ) 
{

    MSCLong32 rv;
    MSCTransmitBuffer tBuffer;

    DPRINTF("%s(%i): MSCUpdateRecord\n",__FILE__,__LINE__);
    
    tBuffer.pBuffer[OFFSET_CLA]    = CLA_00;
    tBuffer.pBuffer[OFFSET_INS]    = INS_UPDATE_RECORD;
    tBuffer.pBuffer[OFFSET_P1]     = index;
    tBuffer.pBuffer[OFFSET_P2]     = 0x04;
    tBuffer.pBuffer[OFFSET_LC]     = dataSize;
 
    memcpy(tBuffer.pBuffer+OFFSET_DATA, pInputData, dataSize);
    tBuffer.bufferSize = dataSize + 5;

    tBuffer.apduResponseSize = MSC_MAXSIZE_BUFFER;
    
    SCARD_RETURN_IF_ERROR( auSCardSendAPDU(pConnection, &tBuffer));
  
    AUTENTHIC_COMMON_COMMAND_RETURN;
}


MSC_RV auDeleteRecord( MSCLPTokenConnection pConnection, MSCUChar8 index)
{
    MSCLong32 rv;
    MSCTransmitBuffer tBuffer;

    DPRINTF("%s(%i): MSCDeleteRecord\n",__FILE__,__LINE__);

    tBuffer.pBuffer[OFFSET_CLA]    = CLA_80;
    tBuffer.pBuffer[OFFSET_INS]    = INS_DELETE_RECORD;
    tBuffer.pBuffer[OFFSET_P1]     = index;
    tBuffer.pBuffer[OFFSET_P2]     = 0x04;
    tBuffer.pBuffer[OFFSET_LC]     = 0x00;
    
    tBuffer.bufferSize = 5;
    tBuffer.apduResponseSize = MSC_MAXSIZE_BUFFER;
    
    SCARD_RETURN_IF_ERROR( auSCardSendAPDU(pConnection, &tBuffer));
   
    AUTENTHIC_COMMON_COMMAND_RETURN;
}



MSC_RV auReadRecord( MSCLPTokenConnection pConnection, 
        MSCUChar8 index, MSCUChar8 size, 
        MSCPUChar8 pOutputData, MSCPUShort16 sizeRecord )
{

    MSCLong32 rv;
    MSCTransmitBuffer tBuffer;

    DPRINTF("%s(%i): MSCReadRecord\n",__FILE__,__LINE__);

    *sizeRecord = 0;
    
    tBuffer.pBuffer[OFFSET_CLA]    = CLA_00;
    tBuffer.pBuffer[OFFSET_INS]    = INS_READ_RECORD;
    tBuffer.pBuffer[OFFSET_P1]     = index;
    tBuffer.pBuffer[OFFSET_P2]     = 0x04;
    tBuffer.pBuffer[OFFSET_LC]     = size;
  
    tBuffer.bufferSize = 5;
    tBuffer.apduResponseSize = MSC_MAXSIZE_BUFFER;
    
    SCARD_RETURN_IF_ERROR( auSCardSendAPDU(pConnection, &tBuffer));
    
    if ( tBuffer.apduResponseSize < 2 )                
        return MSC_UNSPECIFIED_ERROR;
    
    if(tBuffer.apduResponse[tBuffer.apduResponseSize - 2] == 
                AUMSC_SUCCESS_BYTE)  {
        memcpy(pOutputData, tBuffer.apduResponse, size);
        *sizeRecord = size;
        return  MSC_SUCCESS;
    }
  
    AUTENTHIC_COMMON_COMMAND_RETURN;
}



MSC_RV auReadBinary( MSCLPTokenConnection pConnection, 
        MSCULong32 offset, MSCPUChar8 pOutputData, MSCUChar8 dataSize ) 
{
    int ii;
    MSCLong32 rv;
    MSCTransmitBuffer tBuffer;

    DPRINTF("%s(%i): MSCReadBinary: offs 0x%X; size 0x%X\n",
            __FILE__,__LINE__,offset,dataSize);
    
    tBuffer.pBuffer[OFFSET_CLA]    = CLA_00;
    tBuffer.pBuffer[OFFSET_INS]    = INS_READ_BINARY;
    tBuffer.pBuffer[OFFSET_P1]     = (offset & 0x7F00) >> 8;
    tBuffer.pBuffer[OFFSET_P2]     = offset & 0x00FF;
    tBuffer.pBuffer[OFFSET_LC]     = dataSize;

    for (ii=0;ii<2;ii++)   {  
        tBuffer.bufferSize = 5;
        tBuffer.apduResponseSize = MSC_MAXSIZE_BUFFER;
    
        SCARD_RETURN_IF_ERROR( auSCardSendAPDU(pConnection, &tBuffer));
    
        if ( tBuffer.apduResponseSize < 2 )                
            return MSC_UNSPECIFIED_ERROR;
    
        if(tBuffer.apduResponse[tBuffer.apduResponseSize - 2] == 
                AUMSC_SUCCESS_BYTE)  {
            memcpy(pOutputData, tBuffer.apduResponse, dataSize);
            return  MSC_SUCCESS;
        }
        else if (tBuffer.apduResponse[tBuffer.apduResponseSize - 2] == 
                AUMSC_INVALID_LENGTH>>8)  {
            tBuffer.pBuffer[OFFSET_LC] = 
                tBuffer.apduResponse[tBuffer.apduResponseSize - 1];
            continue;
        }
        else
            return convertSW(tBuffer.apduResponse + tBuffer.apduResponseSize - 2);
    }
    
    return convertSW(tBuffer.apduResponse + tBuffer.apduResponseSize - 2);
}


MSC_RV auListFiles( MSCLPTokenConnection pConnection, MSCUChar8 seqOption, 
                       MSCPUShort16 fileID) 
{
    MSCLong32 rv;
    MSCTransmitBuffer tBuffer;
    static int sequenceNumber;
    int obj_num;
    static MSCPUShort16 fileList = NULL;
    static int fileListLength = 0;

    if (fileID!=NULL)
        *fileID = 0xFFFF;
    
    if (seqOption != MSC_SEQUENCE_RESET && fileList !=NULL)   {
        if (sequenceNumber<fileListLength)   {
            DPRINTF("%s(%i): MSCListFiles: seq %i; file %X\n",
                    __FILE__,__LINE__,
                    sequenceNumber, *(fileList+sequenceNumber));
            if (fileID!=NULL)
                *fileID = *(fileList+sequenceNumber);
            
            sequenceNumber++;
            return MSC_SUCCESS;
        }
        else   {
            return MSC_SEQUENCE_END;
        }
    }

    DPRINTF("%s(%i): MSCListFiles\n",__FILE__,__LINE__);
    sequenceNumber = 0;
    if (fileList!=NULL)
        free(fileList);

    fileList=NULL;
    fileListLength = 0;

    tBuffer.pBuffer[OFFSET_CLA]    = CLA_80;
    tBuffer.pBuffer[OFFSET_INS]    = INS_GET_FILE_LIST;
    tBuffer.pBuffer[OFFSET_P1]     = 0x00;
    tBuffer.pBuffer[OFFSET_P2]     = 0x00;
    tBuffer.pBuffer[OFFSET_LC]     = 0x00;
    tBuffer.bufferSize = 5;
    tBuffer.apduResponseSize = MSC_MAXSIZE_BUFFER;
    
    SCARD_RETURN_IF_ERROR( auSCardSendAPDU(pConnection, &tBuffer));
    if ( tBuffer.apduResponseSize < 2 )                
        return MSC_UNSPECIFIED_ERROR;
    
    
    if(tBuffer.apduResponse[tBuffer.apduResponseSize - 2] == AUMSC_SUCCESS_BYTE)  {
        return  MSC_SUCCESS;
    }
    else if (tBuffer.apduResponse[tBuffer.apduResponseSize - 2] == 
            AUMSC_INVALID_LENGTH>>8)  {
        
        fileListLength = tBuffer.apduResponse[tBuffer.apduResponseSize - 1]/2;
        fileList = calloc(fileListLength,sizeof(MSCUShort16));
        if (fileList==NULL)
            return MSC_UNSPECIFIED_ERROR;
        
        tBuffer.pBuffer[OFFSET_CLA]    = CLA_80;
        tBuffer.pBuffer[OFFSET_INS]    = INS_GET_FILE_LIST;
        tBuffer.pBuffer[OFFSET_P1]     = 0x00;
        tBuffer.pBuffer[OFFSET_P2]     = 0x00;
        tBuffer.pBuffer[OFFSET_LC]     = fileListLength * 2;
        tBuffer.bufferSize = 5;
        tBuffer.apduResponseSize = MSC_MAXSIZE_BUFFER;

        SCARD_RETURN_IF_ERROR( auSCardSendAPDU(pConnection, &tBuffer));
        if ( tBuffer.apduResponseSize < 2 )   {
            fileListLength = 0;
            return MSC_UNSPECIFIED_ERROR;
        }
    
        if(tBuffer.apduResponse[tBuffer.apduResponseSize - 2] == AUMSC_SUCCESS_BYTE)  {
            int ii;
            
            for(ii=0; ii<fileListLength; ii++)   
                *(fileList + ii) = tBuffer.apduResponse[2*ii]*0x100 +
                     tBuffer.apduResponse[2*ii + 1];
            
            DPRINTF("%s(%i): MSCListFiles: seq %i; file %X\n",__FILE__,__LINE__,
                     sequenceNumber, *(fileList+sequenceNumber));
            
            if (fileID!=NULL)
                *fileID = *(fileList+sequenceNumber);
            
            sequenceNumber++;
            return  MSC_SUCCESS;
        }
    }
    
    return convertSW(tBuffer.apduResponse + tBuffer.apduResponseSize - 2);
}


MSC_RV auSelectPath( MSCLPTokenConnection pConnection, MSCPUChar8 pathIn)
{
    MSCLong32 rv, id;
    MSCUChar8 response[MSC_MAXSIZE_BUFFER];
    MSCUShort16 responseLen;
    char *path = strdup(pathIn);
    char *beg, *end;
    
    SELECT_MF;
    
    while (strchr(path,'/'))
        *strchr(path,'/') = ' ';
    
    beg = path;

    id = strtol(beg,&end,0x10);
    while (beg!=end)   {
        
        if (id!=0x3F00)   {
            rv = auSelectFile( pConnection, id,
                SELECT_MODE_DF|SELECT_MODE_WITH_ANSWER,
                response, &responseLen);
        }

        if (rv != MSC_SUCCESS)   {
            return MSC_UNSPECIFIED_ERROR;
        }

        beg = end;
        id = strtol(beg,&end,0x10);
    }

    return rv;

}


int TLV_getField(unsigned char *tlv, unsigned char val, unsigned char *ret)
{
    unsigned size, ptr;
    int ii;

    if (*tlv != 0x6F || *(tlv+1)>0x1A)
        return -1;
    
    size = *(tlv+1);

    ptr = 2;

    while (ptr < size && *(tlv+ptr)!=val )   {
        ptr += 2 + *(tlv+ptr+1);
    }

    if (ptr < size)   {
        for (ii=0;ii<*(tlv+ptr+1);ii++)
            *(ret + ii) = *(tlv+ptr+2+ii);
        return 0;
    }
    return -1;
}
 


MSC_RV auListFilesAllocate( MSCLPTokenConnection pConnection,
    MSCPUChar8 path, struct FileInfo **retList)
{
    MSCLong32 rv, id;
    MSCUShort16 ret;
    int ii;
    MSCUChar8 response[MSC_MAXSIZE_BUFFER];
    MSCUShort16 responseLen;
    
    MSCUChar8 tlvType[MSC_MAXSIZE_BUFFER];
    MSCUChar8 tlvSize[MSC_MAXSIZE_BUFFER];
    

    struct FileInfo *files, *ptr;
    files = malloc(sizeof(struct FileInfo));

    files->prev = files;
    files->next = files;
    
    MSC_RETURN_IF_ERROR( auSelectPath(pConnection,path));
    
    rv = auListFiles(pConnection, MSC_SEQUENCE_RESET, &ret);
    for( ;rv == MSC_SUCCESS && ret!=0xFFFF; ) {
        files->next = malloc(sizeof(struct FileInfo));
        files->next->prev = files;
        files = files->next;

        files->id = ret;
        files->path = strdup(path);
        files->list = NULL;
        files->type = 0;
        files->size = 0;
        files->next = files;
        
        rv = auListFiles(pConnection, MSC_SEQUENCE_NEXT, &ret);
    }

    ptr = files;
    
    while (ptr->prev != ptr)   {
        DPRINTF("%s(%i): auListFilesAllocate:: id %X; path %s\n",
            __FILE__,__LINE__, ptr->id, ptr->path);

        rv =  auSelectFile(pConnection,  ptr->id, 
                SELECT_MODE_IMPLICIT | SELECT_MODE_WITH_ANSWER, 
                response, &responseLen); 
        if (responseLen > 2)   {

            if (TLV_getField(response, 0x82, tlvType))  
                return MSC_UNSPECIFIED_ERROR;

            ptr->type = tlvType[0];
            if (ptr->type==0x01)   {
                if (TLV_getField(response, 0x80, tlvSize))  
                    return MSC_UNSPECIFIED_ERROR;
            }
            else   {
                if (TLV_getField(response, 0x85, tlvSize))  
                    return MSC_UNSPECIFIED_ERROR;
            }
            
            if (ptr->type==0x04)   {
                ptr->size = (tlvType[2]*0x100 + tlvType[3]) * tlvType[4];
            }
            else if (ptr->type==0x01)   {
                ptr->size = tlvSize[0]*0x100 + tlvSize[1];
            }
            else if (ptr->type==0x24)   {
                ptr->size = (tlvSize[0]*0x100 + tlvSize[1]) / 8;
            }
            else if (ptr->type==0x2C)   {
                ptr->size = (tlvSize[0]*0x100 + tlvSize[1]) / 8;
            }
            else if (ptr->type==0x38)   {
                char *nextPath = malloc(strlen(path) + 8);
                char *strID = malloc(8);
                sprintf(strID,"/%04X",ptr->id);
                strcat(strcpy(nextPath,path),strID);
                
                auListFilesAllocate( pConnection, nextPath, &ptr->list);
                ptr->size = 0;
            }
        }

        ptr = ptr->prev;
        
        MSC_RETURN_IF_ERROR( auSelectPath(pConnection,path));
    }
    
    *retList = files;
    return rv;
}


MSC_RV auCreateFile( MSCLPTokenConnection pConnection, MSCUChar8 fileType,
        MSCShort16 objectID,	MSCULong32 objectSize, 
        MSCPUChar8 pObjectACL ) 
{
    MSCLong32 rv;
    MSCTransmitBuffer tBuffer;
    MSCPUChar8 apduResponse;
    MSCPUChar8 pBuffer;
    MSCULong32 currentPointer;

    MSCUChar8 acl[8] = {0xFF,0xFF,0xFF,0xFF,
                        0xFF,0xFF,0xFF,0xFF};

    DPRINTF("%s(%i): MSCCreateFile\n",__FILE__,__LINE__);
    pBuffer = tBuffer.pBuffer; 
    apduResponse = tBuffer.apduResponse;

    pBuffer[OFFSET_CLA]    = CLA_00;
    pBuffer[OFFSET_INS]    = INS_CREATE_FILE;
    pBuffer[OFFSET_P1]     = 0x00;
    pBuffer[OFFSET_P2]     = 0x00;
    pBuffer[OFFSET_LC]     = 0x18;
    
    currentPointer = 0;
    pBuffer[OFFSET_DATA + currentPointer++] = 0x62;
    pBuffer[OFFSET_DATA + currentPointer++] = 0x16;
    pBuffer[OFFSET_DATA + currentPointer++] = 0x82;
    pBuffer[OFFSET_DATA + currentPointer++] = 0x02;
    pBuffer[OFFSET_DATA + currentPointer++] = fileType;
    pBuffer[OFFSET_DATA + currentPointer++] = fileType != FILE_TYPE_DF;
    
    pBuffer[OFFSET_DATA + currentPointer++] = 0x83;
    pBuffer[OFFSET_DATA + currentPointer++] = 0x02;
    pBuffer[OFFSET_DATA + currentPointer++] = (objectID>>8) & 0xFF;
    pBuffer[OFFSET_DATA + currentPointer++] = objectID & 0xFF;
    
    pBuffer[OFFSET_DATA + currentPointer++] = 0x85;
    pBuffer[OFFSET_DATA + currentPointer++] = 0x02;
    pBuffer[OFFSET_DATA + currentPointer++] = (objectSize >> 8 ) & 0xFF;
    pBuffer[OFFSET_DATA + currentPointer++] = objectSize  & 0xFF;
    
    pBuffer[OFFSET_DATA + currentPointer++] = 0x86;
    pBuffer[OFFSET_DATA + currentPointer++] = 0x08;
    
    
    memcpy(pBuffer + OFFSET_DATA + currentPointer, pObjectACL, 8);
    
    tBuffer.bufferSize =  OFFSET_DATA + currentPointer + 8;
    tBuffer.apduResponseSize = MSC_MAXSIZE_BUFFER;

    SCARD_RETURN_IF_ERROR( auSCardExchangeAPDU(pConnection, &tBuffer));
    
    strcpy(lastSelectedName,""); 
    
    if ( tBuffer.apduResponseSize != 2 ) 
        return MSC_UNSPECIFIED_ERROR;
        
    if (tBuffer.apduResponse[0] == AUMSC_SUCCESS_BYTE)
        return MSC_SUCCESS;
    else
        return convertSW(apduResponse);
}
/* end of wrappers for the x509 applet functions. */


MSC_RV auReadPublicKey( MSCLPTokenConnection pConnection, MSCUShort16 id,
		       MSCPUChar8 pExp, MSCPUChar8 pMod) 
{
    int ii;
    MSCLong32 rv;
    MSCTransmitBuffer tBuffer;

    DPRINTF("%s(%i): read_public_key\n",__FILE__,__LINE__);
    
    tBuffer.pBuffer[OFFSET_CLA]    = CLA_80;
    tBuffer.pBuffer[OFFSET_INS]    = INS_READ_PUBLIC;
    tBuffer.pBuffer[OFFSET_P1]     = 0x01;
    tBuffer.pBuffer[OFFSET_P2]     = 0x04;
    tBuffer.pBuffer[OFFSET_LC]     = MSC_EXP_SIZE;

    for (ii=0;ii<2;ii++)   {  
        tBuffer.bufferSize = 5;
        tBuffer.apduResponseSize = MSC_MAXSIZE_BUFFER;
    
        SCARD_RETURN_IF_ERROR( auSCardSendAPDU(pConnection, &tBuffer));
    
        if ( tBuffer.apduResponseSize < 2 )                
            return MSC_UNSPECIFIED_ERROR;
    
        if(tBuffer.apduResponse[tBuffer.apduResponseSize - 2] == 
                AUMSC_SUCCESS_BYTE)  {
            memcpy(pExp, tBuffer.apduResponse, MSC_EXP_SIZE);
            break;
        }
        else if (tBuffer.apduResponse[tBuffer.apduResponseSize - 2] == 
                AUMSC_INVALID_LENGTH>>8)  {
            tBuffer.pBuffer[OFFSET_LC] = 
                tBuffer.apduResponse[tBuffer.apduResponseSize - 1];
            continue;
        }
        else
            return convertSW(tBuffer.apduResponse + tBuffer.apduResponseSize - 2);
    }
    
    tBuffer.pBuffer[OFFSET_P1]     = 0x02;
    tBuffer.pBuffer[OFFSET_LC]     = MSC_MOD_SIZE;
    
    for (ii=0;ii<2;ii++)   {  
        tBuffer.bufferSize = 5;
        tBuffer.apduResponseSize = MSC_MAXSIZE_BUFFER;
    
        SCARD_RETURN_IF_ERROR( auSCardSendAPDU(pConnection, &tBuffer));
    
        if ( tBuffer.apduResponseSize < 2 )                
            return MSC_UNSPECIFIED_ERROR;
    
        if(tBuffer.apduResponse[tBuffer.apduResponseSize - 2] == 
                AUMSC_SUCCESS_BYTE)  {
            memcpy(pMod, tBuffer.apduResponse, MSC_MOD_SIZE);
            return  MSC_SUCCESS;
        }
        else if (tBuffer.apduResponse[tBuffer.apduResponseSize - 2] == 
                AUMSC_INVALID_LENGTH>>8)  {
            tBuffer.pBuffer[OFFSET_LC] = 
                tBuffer.apduResponse[tBuffer.apduResponseSize - 1];
            continue;
        }
        else
            return convertSW(tBuffer.apduResponse + tBuffer.apduResponseSize - 2);
    }

    return convertSW(tBuffer.apduResponse + tBuffer.apduResponseSize - 2);
}



MSC_RV auUpdateKeyComponent( MSCLPTokenConnection pConnection, 
        key_component_t compType, MSCPUChar8 compData, MSCUShort16 compLen) 
{
    int ii;
    MSCLong32 rv;
    MSCTransmitBuffer tBuffer;

    DPRINTF("%s(%i): update_key_comp: type %i; len %i\n",
            __FILE__,__LINE__, compType, compLen);
    
    tBuffer.pBuffer[OFFSET_CLA]    = CLA_80;
    tBuffer.pBuffer[OFFSET_INS]    = INS_UPDATE_KEY;
    tBuffer.pBuffer[OFFSET_P1]     = compType;
    tBuffer.pBuffer[OFFSET_P2]     = 0x04;
    tBuffer.pBuffer[OFFSET_LC]     = compLen;

    memcpy(tBuffer.pBuffer+OFFSET_DATA, compData, compLen);
    tBuffer.bufferSize = compLen + 5;

    tBuffer.apduResponseSize = MSC_MAXSIZE_BUFFER;
    
    SCARD_RETURN_IF_ERROR( auSCardSendAPDU(pConnection, &tBuffer));
    
    AUTENTHIC_COMMON_COMMAND_RETURN;
}

/* Get object file info. */
MSC_RV PL_MSCGetObjectAttributes( MSCLPTokenConnection pConnection,
			       MSCString objectID, MSCLPObjectInfo pObjectInfo ) 
{
    MSC_RV rv;
    int index;
    MSCUChar8 response[MSC_MAXSIZE_BUFFER];
    MSCUShort16 responseLen;
    MSCLPObjectInfo p_objInfo = (MSCLPObjectInfo) &response[0];

    DPRINTF("%s(%i): MSCGetObjectAttributes\n",__FILE__,__LINE__);
    if ( pConnection == NULL ) 
        return MSC_INVALID_PARAMETER; 

    /* Get first record. */
    MSC_RETURN_IF_ERROR( PL_MSCListObjects( pConnection, 0, p_objInfo));

    for (index=1; strcmp(p_objInfo->objectID, objectID)!=0; index++)   {
        DPRINTF("%s(%i): MSCGetObjectAttributes\n",__FILE__,__LINE__);
        rv = auReadRecord( pConnection, 
                index, OBJECT_INFO_SIZE, 
                &response[0], &responseLen );
        if (rv != MSC_SUCCESS)
            break;
    }

    if(rv==MSC_SUCCESS)   {
        memcpy(&pObjectInfo->objectACL, &p_objInfo->objectACL,
                sizeof(pObjectInfo->objectACL));
        pObjectInfo->objectSize = p_objInfo->objectSize;
        strncpy(pObjectInfo->objectID, p_objInfo->objectID, MSC_MAXSIZE_OBJID);
        return MSC_SUCCESS;  
    }
    else
        return MSC_OBJECT_NOT_FOUND;
}


/*
 * Import the RSA key to smart card.
 * It overwrite key file, if exists.
 * pKeyBlob holds the pointer to RSA key.
 * The RSA key component to import the key (public or private) 
 * defined with pAddParams and addParamsSize. 
 */
MSC_RV auImportRSAKey( MSCLPTokenConnection pConnection, MSCUChar8 keyNum,
		     MSCLPKeyACL pKeyACL, RSA *rsa, MSCLPKeyPolicy keyPolicy,
		     MSCUChar8 keyType, MSCUChar8 keyPartner ) 
{
    MSCLong32 rv;
    MSCKeyInfo keyStruct;
    MSCUChar8 response[MSC_MAXSIZE_BUFFER];
    MSCUShort16 responseLen;
    MSCUChar8 pExp[MSC_EXP_SIZE];
    MSCUChar8 pMod[MSC_MOD_SIZE];
    unsigned char hex[0x100], hex_comp[0x100];
    int ii;
    key_component_t comp;
    MSCKeyACL privACL;
    MSCKeyACL pubACL;

    DPRINTF("%s(%i): auImportRSAKey\n",__FILE__,__LINE__);

    /*
    if(BN_num_bits(rsa->n) != CF_1024_KEYSIZE)   {
        DPRINTF("MSCImportRSAKey: Invalid rsa key size: %u\n",BN_num_bits(rsa->n));
        return MSC_INVALID_PARAMETER;
    }
    */
    /* Select key container. */
    SELECT_MF;

    MSC_RETURN_IF_ERROR( auSelectFile( pConnection, AUTHENTIC_DIR_ID,
            SELECT_MODE_DF|SELECT_MODE_WITH_ANSWER,
            response, &responseLen));
    
    MSC_RETURN_IF_ERROR( auSelectFile( pConnection, AUTHENTIC_KEYDIR_ID,
            SELECT_MODE_DF|SELECT_MODE_WITH_ANSWER,
            response, &responseLen));
    
    rv = auSelectFile( pConnection, AUTHENTIC_KEY_ID + keyNum,
            SELECT_MODE_EF|SELECT_MODE_WITH_ANSWER,
            response, &responseLen);
    
    /* Delete if it exists. */
    if (rv == MSC_SUCCESS)   
        MSC_RETURN_IF_ERROR( auDeleteFile ( pConnection,
                    AUTHENTIC_KEY_ID + keyNum, FILE_TYPE_EF_TRANSPARENT));
    
    
    if(keyType == MSC_KEY_RSA_PUBLIC)   {
        DPRINTF("%s(%i): auImportRSAKey\n",__FILE__,__LINE__);
        /* 
         * Create public key container and update it's components.
         */
        MSC_RETURN_IF_ERROR( auCreateFile(pConnection, FILE_TYPE_RSA_PUBLIC, 
                    AUTHENTIC_KEY_ID + keyNum, 0x0400, publicKeyACL));

        ii = BN_bn2bin(rsa->e,hex_comp);
        comp = e;
        MSC_RETURN_IF_ERROR( auUpdateKeyComponent( pConnection, comp, 
                    hex_comp, ii));

        ii = BN_bn2bin(rsa->n,hex);
        memset(hex_comp,0,0x80);
        memcpy(hex_comp+0x80-ii,hex,ii);
        comp = n;
            
        DPRINTF("%s(%i): auImportRSAKey\n",__FILE__,__LINE__);
        MSC_RETURN_IF_ERROR( auUpdateKeyComponent( pConnection, comp, 
                    hex, 0x80));
            
        pubACL.readPermission  = MSC_AUT_ALL;
        pubACL.writePermission = MSC_AUT_PIN_1;
        pubACL.usePermission   = MSC_AUT_ALL;
        MSC_RETURN_IF_ERROR( auWriteKeyInfo( pConnection, 
                    keyNum, 
                    MSC_KEY_RSA_PUBLIC, 
                    keyPartner, 
                    CF_1024_KEYSIZE, 
                    keyPolicy,
                    &pubACL));
    }
    else if (keyType == MSC_KEY_RSA_PRIVATE_CRT)   {
        /* 
         * Create private key container and update it's components.
         */
        DPRINTF("%s(%i): auImportRSAKey\n",__FILE__,__LINE__);
        MSC_RETURN_IF_ERROR( auCreateFile(pConnection, 
                    FILE_TYPE_RSA_PRIVATE,
                    AUTHENTIC_KEY_ID + keyNum, 
                    RSA_1024_SIZE, 
                    privateKeyACL));
    
        ii = BN_bn2bin(rsa->p,hex);
        memset(hex_comp, 0, RSA_1024_COMPSIZE);
        memcpy(hex_comp+RSA_1024_COMPSIZE-ii, hex, ii);
        comp = p;
        MSC_RETURN_IF_ERROR( auUpdateKeyComponent( pConnection, 
                    comp, 
                    hex_comp, 
                    RSA_1024_COMPSIZE));
            
        ii = BN_bn2bin(rsa->q,hex);
        memset(hex_comp, 0, RSA_1024_COMPSIZE);
        memcpy(hex_comp+RSA_1024_COMPSIZE-ii, hex, ii);
        comp = q;
        MSC_RETURN_IF_ERROR( auUpdateKeyComponent( pConnection, 
                    comp, 
                    hex_comp, 
                    RSA_1024_COMPSIZE));
        
        ii = BN_bn2bin(rsa->iqmp,hex);
        memset(hex_comp, 0, RSA_1024_COMPSIZE);
        memcpy(hex_comp+RSA_1024_COMPSIZE-ii, hex, ii);
        comp = iqmp;
        MSC_RETURN_IF_ERROR( auUpdateKeyComponent( pConnection, 
                    comp, 
                    hex_comp, 
                    RSA_1024_COMPSIZE));
        
        ii = BN_bn2bin(rsa->dmp1,hex);
        memset(hex_comp, 0, RSA_1024_COMPSIZE);
        memcpy(hex_comp+RSA_1024_COMPSIZE-ii, hex, ii);
        comp = dmp1;
        MSC_RETURN_IF_ERROR( auUpdateKeyComponent( pConnection, 
                    comp, 
                    hex_comp, 
                    RSA_1024_COMPSIZE));
            
        ii = BN_bn2bin(rsa->dmq1,hex);
        memset(hex_comp, 0, RSA_1024_COMPSIZE);
        memcpy(hex_comp+RSA_1024_COMPSIZE-ii, hex, ii);
        comp = dmq1;
        MSC_RETURN_IF_ERROR( auUpdateKeyComponent( pConnection, 
                    comp, 
                    hex_comp, 
                    RSA_1024_COMPSIZE));
            
        privACL.readPermission  = MSC_AUT_NONE;
        privACL.writePermission = MSC_AUT_PIN_1;
        privACL.usePermission   = MSC_AUT_PIN_1;
        MSC_RETURN_IF_ERROR( auWriteKeyInfo( pConnection, 
                    keyNum, 
                    MSC_KEY_RSA_PRIVATE_CRT, 
                    keyPartner, 
                    CF_1024_KEYSIZE, 
                    keyPolicy,
                    &privACL));
    }
    else   {
        DPRINTF("%s(%i): auImportRSAKey: Invalid file type: %X\n",
             __FILE__,__LINE__, keyType);
        rv = MSC_UNSUPPORTED_FEATURE;
    }

    return rv;
}
 


MSC_RV auUnblockPIN( MSCLPTokenConnection pConnection, 
        MSCUChar8 pinNumber,
		MSCPUChar8 pUnblockCode, MSCULong32 unblockCodeSize,
        MSCPUChar8 pNewUserPIN, MSCULong32 newUserPINSize) 
{
    MSCLong32 rv;
    MSCTransmitBuffer tBuffer;
    MSCPUChar8 apduResponse;
    MSCPUChar8 pBuffer;
    MSCUChar8 pinN;
    int ii;

    DPRINTF("%s(%i): MSCUnblockPIN %X\n",__FILE__,__LINE__, pinNumber);
    if ( (pUnblockCode==NULL || unblockCodeSize > 15 ||  unblockCodeSize <0) ||
         pinNumber != NUM_PIN )   {
        DPRINTF("%s(%i):Invalid Parameter: pinNum %d; codeSize: %d\n",
                __FILE__,__LINE__, pinNumber, unblockCodeSize);
        return MSC_INVALID_PARAMETER;
    }

    pinN = authentICPins[pinNumber];

    if (pNewUserPIN==NULL)
        newUserPINSize = 0;

    if (newUserPINSize!=0 && newUserPINSize<4 && newUserPINSize>15)   {
        DPRINTF("%s(%i):Invalid Parameter: userPINSize %d\n",
                __FILE__,__LINE__, newUserPINSize);
        return MSC_INVALID_PARAMETER;
    }
        
    
    pBuffer = tBuffer.pBuffer; 
    apduResponse = tBuffer.apduResponse;

    pBuffer[OFFSET_CLA]    = CLA_00;
    pBuffer[OFFSET_INS]    = INS_RESET_COUNTER;
    
    pBuffer[OFFSET_P2]     = pinN;

    tBuffer.apduResponseSize = MSC_MAXSIZE_BUFFER;
    
    memcpy(&pBuffer[OFFSET_DATA], pUnblockCode, unblockCodeSize);
    for (ii = unblockCodeSize;ii<0x0F;ii++)
        pBuffer[OFFSET_DATA+ii] = 0xFF;
    
    if (newUserPINSize)   {
        pBuffer[OFFSET_LC]     = 0x1E;
        pBuffer[OFFSET_P1] = 0x00;
        tBuffer.bufferSize = 0x23;
        memcpy(&pBuffer[OFFSET_DATA + 0x0F], pNewUserPIN, newUserPINSize);
        for (ii = newUserPINSize;ii<0x0F;ii++)
            pBuffer[OFFSET_DATA + 0x0F + ii] = 0xFF;
    }
    else   {
        pBuffer[OFFSET_LC]     = 0x0F;
        pBuffer[OFFSET_P1] = 0x01;
        tBuffer.bufferSize = 0x14;
    }


    tBuffer.apduResponseSize = MSC_MAXSIZE_BUFFER;

    SCARD_RETURN_IF_ERROR( auSCardSendAPDU(pConnection, &tBuffer));

    AUTENTHIC_COMMON_COMMAND_RETURN;
}


MSCLong32 auSCardSendAPDU( MSCLPTokenConnection pConnection, 
			     MSCLPTransmitBuffer tBuffer) {
  
    int i;
    MSCLong32 rv, ret;
    MSCULong32 originalLength;
    MSCULong32 dwActiveProtocol;

    originalLength = tBuffer->apduResponseSize;

#ifdef MSC_DEBUG
    INSPrintDebug(tBuffer->pBuffer[OFFSET_INS]);
    
    DPRINTF("\n->: ");
    for (i=0; i < tBuffer->bufferSize; i++) 
        DPRINTF("%02X ", tBuffer->pBuffer[i]);
    DPRINTF("\n");
#endif
    
    while(1) {
        tBuffer->apduResponseSize = originalLength;
      
        rv =  SCardTransmit(pConnection->hCard, pConnection->ioType,
			    tBuffer->pBuffer, tBuffer->bufferSize, 0,
			    tBuffer->apduResponse, &tBuffer->apduResponseSize );
      
        if ( rv == SCARD_S_SUCCESS ) {
	        break;
	    } 
        else if ( rv == SCARD_W_RESET_CARD ) {
	        pConnection->tokenInfo.tokenType |= MSC_TOKEN_TYPE_RESET;
	        
            ret = SCardReconnect(pConnection->hCard, 
                            pConnection->shareMode, 
			                SCARD_PROTOCOL_T0 | SCARD_PROTOCOL_T1,
			                SCARD_LEAVE_CARD, &dwActiveProtocol );
	
	        PL_MSCIdentifyToken(pConnection);
	
	        if ( ret == SCARD_S_SUCCESS ) 
	            continue;
        } 
        else if ( rv == SCARD_W_REMOVED_CARD ) {
	        /* Push REMOVED_TOKEN back to the application */
	        pConnection->tokenInfo.tokenType = MSC_TOKEN_TYPE_REMOVED;
	        return rv;
        } 
        else {
	        /* Must be a BIG, BAD, ERROR */

	        DPRINTF("Transmit error: %s\n", pcsc_stringify_error(rv));

	        return rv;
        }
    }
    
#ifdef MSC_DEBUG
    DPRINTF("<-: ");
    
    for (i=0; i < tBuffer->apduResponseSize; i++) 
        DPRINTF("%02X ", tBuffer->apduResponse[i]);
    DPRINTF("\n");
#endif

  return rv;
}


MSCLong32 auSCardExchangeAPDU( MSCLPTokenConnection pConnection, 
			     MSCLPTransmitBuffer tBuffer ) 
{
    MSCLong32 rv, ret;
   
    DPRINTF("%s(%i): auSCardExchangeAPDU\n",__FILE__,__LINE__);
    
    rv = auSCardSendAPDU (pConnection, tBuffer );
    
    if(tBuffer->apduResponseSize == 2 && 
            tBuffer->apduResponse[0] == AUMSC_CORRECT_EXECUTION)   {
        tBuffer->pBuffer[OFFSET_CLA] = CLA_00;
        tBuffer->pBuffer[OFFSET_INS] = INS_GET_RESPONSE;
        tBuffer->pBuffer[OFFSET_P1]  = 0x00;
        tBuffer->pBuffer[OFFSET_P2]  = 0x00;
        tBuffer->pBuffer[OFFSET_LC]  = tBuffer->apduResponse[1];
 
        tBuffer->bufferSize = 5;
        tBuffer->apduResponseSize = MSC_MAXSIZE_BUFFER;
        
        rv = auSCardSendAPDU(pConnection, tBuffer);
    }

    return rv;
}



MSC_RV pcscToMSC( MSCLong32 pcscCode ) 
{

    switch(pcscCode) {
        case SCARD_S_SUCCESS:
            return MSC_SUCCESS;
            
        case SCARD_E_INVALID_HANDLE:
            return MSC_INVALID_HANDLE;
        case SCARD_E_SHARING_VIOLATION:
            return MSC_SHARING_VIOLATION;
        case SCARD_W_REMOVED_CARD:
            return MSC_TOKEN_REMOVED;
        case SCARD_E_NO_SMARTCARD:
            return MSC_TOKEN_REMOVED;
        case SCARD_W_RESET_CARD:
            return MSC_TOKEN_RESET;
        case SCARD_W_INSERTED_CARD:
            return MSC_TOKEN_INSERTED;
        case SCARD_E_NO_SERVICE:
            return MSC_SERVICE_UNRESPONSIVE;
        case SCARD_E_UNKNOWN_CARD:
        case SCARD_W_UNSUPPORTED_CARD:
        case SCARD_E_CARD_UNSUPPORTED:
            return MSC_UNRECOGNIZED_TOKEN;
        case SCARD_E_INVALID_PARAMETER:
        case SCARD_E_INVALID_VALUE:
        case SCARD_E_UNKNOWN_READER:
        case SCARD_E_PROTO_MISMATCH:
        case SCARD_E_READER_UNAVAILABLE:
            return MSC_INVALID_PARAMETER;
        case SCARD_E_CANCELLED:
            return MSC_CANCELLED;
        case SCARD_E_TIMEOUT:
            return MSC_TIMEOUT_OCCURRED;

        default:
            return MSC_INTERNAL_ERROR;
    }
}


MSCUShort16 convertSW(MSCPUChar8 pBuffer) 
{
    MSCUShort16 retValue;
    MSCUShort16 newValue;

    retValue  = pBuffer[0] * 0x100;
    retValue += pBuffer[1];

    switch(retValue) {
        case AUMSC_SUCCESS:
            newValue = MSC_SUCCESS;
            break;
        case AUMSC_CARD_BLOCKED:
            newValue = MSC_OPERATION_NOT_ALLOWED;
            break;
        case AUMSC_AUTH_FAILED:
            newValue = MSC_AUTH_FAILED;
            break;
        case AUMSC_INCORRECT_DATA_LENGTH:
        case AUMSC_NOT_INITIALIZED_KEY:
        case AUMSC_INCORRECT_START_OFFSET:
        case AUMSC_INVALID_LENGTH:
        case AUMSC_INCORRECT_CLA:
            newValue = MSC_INVALID_PARAMETER;
            break;
        case AUMSC_SECURITY_VIOLATION:
            newValue = MSC_UNAUTHORIZED;
            break;
        case AUMSC_OPERATION_NOT_ALLOWED:
            newValue = MSC_OPERATION_NOT_ALLOWED;
            break;
        case AUMSC_CONDITION_NOT_SATISFIED:
            newValue = MSC_OPERATION_NOT_ALLOWED;
            break;
        case AUMSC_WRONG_FILE_TYPE:
        case AUMSC_NO_CURRENT_EF:
        case AUMSC_INCORRECT_TAG:
        case AUMSC_WRONG_EXPONENET_SIZE:
            newValue = MSC_OPERATION_NOT_ALLOWED;
            break;
        case AUMSC_NO_MEMORY_LEFT:
            newValue = MSC_NO_MEMORY_LEFT;
            break;
        case AUMSC_OBJECT_NOT_FOUND:
        case AUMSC_RECORD_NOT_FOUND:  
            newValue = MSC_OBJECT_NOT_FOUND;
            break;
        case AUMSC_INCORRECT_P1P2:
            newValue = MSC_INCORRECT_P1;
            break;
        case AUMSC_UNSUPPORTED_FEATURE:
            newValue = MSC_UNSUPPORTED_FEATURE;
            break;
        case AUMSC_FILE_EXIST:
            newValue = MSC_OBJECT_EXISTS;
            break;
    
        default:
            newValue = retValue;
            break;
    }

    return newValue;
}



void INSPrintDebug(MSCUChar8 ch)
{
#ifdef MSC_DEBUG
#define DEBUG_INS(a, b) if ((a) == (b)) DPRINTF("[" #b "] ");
#else
#define DEBUG_INS(a, b)
#endif
    DEBUG_INS(ch, INS_SELECT_APPLICATION);
    DEBUG_INS(ch, INS_GET_CHALLENGE);
    DEBUG_INS(ch, INS_REFERENCE);
    DEBUG_INS(ch, INS_RESET_COUNTER);
    DEBUG_INS(ch, INS_VERIFY_PIN);
    DEBUG_INS(ch, INS_CREATE_FILE);
    DEBUG_INS(ch, INS_GET_FILE_LIST);
    DEBUG_INS(ch, INS_SELECT_FILE);
    DEBUG_INS(ch, INS_DELETE_FILE);
    DEBUG_INS(ch, INS_EXTEND_BINARY);
    DEBUG_INS(ch, INS_READ_BINARY);
    DEBUG_INS(ch, INS_UPDATE_BINARY);
    DEBUG_INS(ch, INS_APPEND_RECOD);
    DEBUG_INS(ch, INS_DELETE_RECORD);
    DEBUG_INS(ch, INS_READ_RECORD);
    DEBUG_INS(ch, INS_UPDATE_RECORD);
    DEBUG_INS(ch, INS_READ_PUBLIC);
    DEBUG_INS(ch, INS_UPDATE_KEY);
    DEBUG_INS(ch, INS_SIGNATURE_MSE);
    DEBUG_INS(ch, INS_SIGNATURE_PSO);
    DEBUG_INS(ch, INS_GENERATE_KEY);
    DEBUG_INS(ch, INS_GET_RESPONSE);
    DEBUG_INS(ch, INS_GET_DATA);
}



MSC_RV getObjectID( MSCLPTokenConnection pConnection,  MSCPUShort16 objectID, 
        MSCString objectString, MSCUShort16 valid_name) 
{
    MSCLong32 rv;	
    MSCUShort16 id;
    MSCUChar8  response[MAX_BUFFER_SIZE];
    MSCUShort16 responseLen;
    int obj, jj;
  
    DPRINTF("%s(%i): objectString %s; valid_name %i\n",
            __FILE__,__LINE__,objectString,valid_name);
    
    if (!valid_name) 
        SELECT_MF;
            
    for( obj=0; obj<NN_OBJECTS; obj++)   
        if(strcmp(objects[obj].name,objectString)==0)
            break;
    
    if (obj<NN_OBJECTS)   {
            
        for (jj=0;objects[obj].dir[jj]!=0;jj++)   {
                DPRINTF("%s(%i): select %X\n",
                                    __FILE__,__LINE__, objects[obj].dir[jj]);
            if(!valid_name)   
                MSC_RETURN_IF_ERROR( auSelectFile( pConnection, 
                            objects[obj].dir[jj],
                            SELECT_MODE_DF|SELECT_MODE_WITH_ANSWER,
                            response, &responseLen));
        }
    
        id = objects[obj].id;
    }

    else if (strlen(objectString)==2)   {

        if (!valid_name)  {
            MSC_RETURN_IF_ERROR( auSelectFile( pConnection, AUTHENTIC_DIR_ID,
                    SELECT_MODE_DF|SELECT_MODE_WITH_ANSWER,
                    response, &responseLen));
        
            MSC_RETURN_IF_ERROR( auSelectFile( pConnection, AUTHENTIC_OBJECTDIR_ID,
                    SELECT_MODE_DF|SELECT_MODE_WITH_ANSWER,
                    response, &responseLen));
        }

        id = *((MSCPUShort16)objectString);
        DPRINTF("%s:%i id %04X\n",__FILE__,__LINE__,id);
    }
    
    else   {
        char *ptr = objectString;
        char id_str[6];
        
        errno = 0;
        
        if (strlen(ptr)/4 > 1)   {

            for(;strlen(ptr)/4 > 1; ptr+=4)   {
                strncpy(id_str,ptr,4);
                *(id_str+4) = '\0';
                errno = 0;
                id = strtol(id_str,NULL,0x10);
                if (errno==EINVAL)
                    return MSC_INVALID_PARAMETER;
                
                if(!valid_name)   {
                    rv = auSelectFile( pConnection, id,
                            SELECT_MODE_DF|SELECT_MODE_WITH_ANSWER,
                            response, &responseLen);
                    if ( rv != MSC_SUCCESS ) return rv;
                }
            }
            
            id = strtol(ptr, NULL, 0x10);
            if (errno==EINVAL)
                return MSC_INVALID_PARAMETER;
        }
    }
   
    *objectID = id;

    return MSC_SUCCESS;
}


//void MemCopy16(MSCPUChar8 destValue, MSCPUShort16 srcValue) 
void MemCopy16(void *outValue, MSCPUShort16 srcValue) 
{
    MSCPUChar8 destValue = (MSCPUChar8)outValue;
    
    destValue[0] = (*srcValue & 0xFF00) >> 8;
    destValue[1] = (*srcValue & 0x00FF);
}


void MemCopyTo16(MSCPUShort16 destValue, void *inValue) 
{
    MSCPUChar8 srcValue = (MSCPUChar8)inValue;

    *destValue  = srcValue[0] * 0x100;
    *destValue += srcValue[1];
}


void autentIC_Log(char *msg, ...)
{
    va_list args;
    char format[0x200];

    FILE *fp;
    fp = fopen(autentIC_LogFilename, "ab");

    if (!fp)    {
        fp = stderr;
        fprintf(fp, "Error, could not open: %s\n", autentIC_LogFilename);
    }

    sprintf(format, "%s", msg);
    va_start(args, msg);
    vfprintf(fp, format, args);
    va_end(args);

    fflush(fp);

    if (fp != stderr)
        fclose(fp);
}

void getOberthurACL(MSCPUChar8 out_acl, MSCPUChar8 in_acl)
{
    MSCLPObjectACL acl = (MSCLPObjectACL)in_acl;
    MSCUShort16 ii;
   
    memset(out_acl,0xFF,8);

    if (acl->readPermission==0)
        *(out_acl+2) = 0x0F;
    else if (acl->readPermission!=0xFFFF)
        *(out_acl+2) = 0x4F;

    
    if (acl->writePermission==0)
        *(out_acl+0) = 0x0F, *(out_acl+1) = 0x0F;
    else if (acl->writePermission!=0xFFFF)
        *(out_acl+0) = 0x4F, *(out_acl+1) = 0x4F;
        
//TODO  
//For a moment we do not want create the object for which
//the delete operation is never possible.
     *(out_acl+3) = 0x4F;    
    
    if (acl->deletePermission==0)
        *(out_acl+3) = 0x0F;
    else if (acl->deletePermission!=0xFFFF)
        *(out_acl+3) = 0x4F;
    
}
