/******************************************************************
 
        MUSCLE SmartCard Development ( http://www.musclecard.com )
            Title  : authentIC.h
            Package: authentIC
            Author : David Corcoran
            Date   : 02/19/02
            License: Copyright (C) 2002 David Corcoran
                     <corcoran@linuxnet.com>
            Purpose: To provide a generic DDK for plug-ins.
 
 
********************************************************************/
/********************************************************
 * PlugIn for Oberthur AuthentIC with applet X509
 *
 * Copyright (C) 2002 IDEALX (http://www.idealx.com/).
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of either the GNU General Public License
 * or the Clarified Artistic License, as specified in the LICENSE
 * file.
 *
 * Initial implementaion by Victor Tarasov vtarasov@idealx.com
 * ********************************************************/


#ifndef __authentIC_h__
#define __authentIC_h__

#include <openssl/rsa.h>
#include <openssl/dsa.h>

#ifdef __cplusplus
  extern "C" {
#endif
	  
#define CLA_00          0x00
#define CLA_80          0x80
	  
/* Some useful offsets in the buffer */
#define OFFSET_CLA	0x00
#define OFFSET_INS	0x01
#define OFFSET_P1	0x02
#define OFFSET_P2	0x03
#define OFFSET_LC	0x04
#define OFFSET_DATA	0x05

#define OFFSET_ENCODING 0x00
#define OFFSET_KEYTYPE  0x01
#define OFFSET_KEYSIZE  0x02
#define OFFSET_KEYDATA  0x04

#define RSA_1024_SIZE       0x400
#define RSA_1024_COMPSIZE   0x40

#define MC_1024P_FULLSIZE     140
#define MC_SIZEOF_COMPSIZE      2
#define MC_1024_OFFSET_P        4
#define MC_1024_OFFSET_Q       70
#define MC_1024_OFFSET_PQ     136
#define MC_1024_OFFSET_DP1    202
#define MC_1024_OFFSET_DQ1    268


#define MC_1024P_MOD            4
#define MC_1024P_EXP          134
#define MC_1024P_MODSIZE      128

#define INS_SELECT_APPLICATION 	0xA4
#define INS_GET_CHALLENGE 	0x84
#define INS_REFERENCE		0x24
#define INS_RESET_COUNTER	0x2C
#define INS_VERIFY_PIN		0x20
	 
#define INS_CREATE_FILE 	0xE0
#define INS_GET_FILE_LIST	0x34
#define INS_SELECT_FILE		0xA4
#define INS_DELETE_FILE		0xE4
#define INS_EXTEND_BINARY	0xD4
#define INS_READ_BINARY		0xB0
#define INS_UPDATE_BINARY	0xD6
#define INS_APPEND_RECOD	0xE2
#define INS_DELETE_RECORD	0x32
#define INS_READ_RECORD		0xB2
#define INS_UPDATE_RECORD	0xDC
	  
#define INS_READ_PUBLIC		0xB2
#define INS_UPDATE_KEY		0xDC
#define INS_SIGNATURE_MSE	0x22
#define INS_SIGNATURE_PSO	0x2A
#define INS_GENERATE_KEY	0x46
#define INS_GET_RESPONSE	0xC0
#define INS_GET_DATA		0xCA

typedef enum key_component {
    e = 1, n, d=0, p, q, iqmp, dmp1, dmq1
} key_component_t;
    
#define SELECT_MODE_IMPLICIT	0x00
#define SELECT_MODE_DF		0x01
#define SELECT_MODE_EF		0x02
#define SELECT_MODE_PARENT	0x03

#define SELECT_MODE_NO_ANSWER	0x0C
#define SELECT_MODE_WITH_ANSWER	0x00

#define SO_PIN      0x00
#define USER_PIN    0x80
#define NUM_SOPIN	0
#define NUM_PIN		1

#define FILE_TYPE_DF			    0x38
#define FILE_TYPE_EF_TRANSPARENT	0x01
#define FILE_TYPE_EF_VARIABLE		0x04
#define FILE_TYPE_RSA_PUBLIC		0x24
#define FILE_TYPE_RSA_PRIVATE		0x2C
	  
	  
#define AUMSC_CORRECT_EXECUTION		0x61
	  
#define AUMSC_CARD_BLOCKED		0x6283
#define AUMSC_AUTH_FAILED		0x63CF
#define AUMSC_INCORRECT_DATA_LENGTH	0x6700
	  
#define AUMSC_WRONG_FILE_TYPE		0x6981 	//Check the meaning for  RSA keys:
#define AUMSC_SECURITY_VIOLATION	0x6982
#define AUMSC_OPERATION_NOT_ALLOWED	0x6983
#define AUMSC_CONDITION_NOT_SATISFIED	0x6985
#define AUMSC_NO_CURRENT_EF		0x6986
	  
#define AUMSC_INCORRECT_TAG		        0x6A80	
                        //for RSA:
 					    //exponent is not odd

#define AUMSC_WRONG_EXPONENET_SIZE	    0x6A81	  
#define AUMSC_OBJECT_NOT_FOUND		    0x6A82
#define AUMSC_RECORD_NOT_FOUND		    0x6A83
#define AUMSC_NO_MEMORY_LEFT	    0x6A84	//for MSE signature: 
						//not RSA private key file
#define AUMSC_INCORRECT_P1P2		0x6A86
#define AUMSC_NOT_INITIALIZED_KEY	0x6A88 
#define AUMSC_FILE_EXIST		    0x6A89
	  
#define AUMSC_INCORRECT_START_OFFSET	0x6B00
#define AUMSC_INVALID_LENGTH		0x6CFF
#define AUMSC_UNSUPPORTED_FEATURE	0x6D00
#define AUMSC_INCORRECT_CLA		0x6E00

	  
#define AUMSC_SUCCESS		0x9000
#define AUMSC_SUCCESS_BYTE	(AUMSC_SUCCESS>>8)


#define AUTHENTIC_RSA_KEYSTART      -1
#define AUTHENTIC_DES_KEYSTART      1

#define AUTHENTIC_DIR_ID            0x3FAB
//#define AUTHENTIC_DIR_ID            0x3FAA

#define AUTHENTIC_KEYDIR_ID         0x3FCE
#define AUTHENTIC_OBJECTDIR_ID      0x3FCF
#define AUTHENTIC_KEY_ID            0x1000

#define AUTHENTIC_KEYINFO_ID        0xF301      
#define str_AUTHENTIC_KEYINFO_ID    "0xF301"

#define AUTHENTIC_USERINFO_ID       0xF302
#define str_AUTHENTIC_USERINFO_ID   "0xF302"
#define  AUTHENTIC_USERINFO_LENGTH  0xFF

#define AUTHENTIC_OBJECTINFO_ID       0xF401
#define str_AUTHENTIC_OBJECTINFO_ID   "0xF401"
#define  OBJECT_INFO_SIZE  (sizeof(MSCObjectInfo))
#define MAX_NUMBER_OBJECT_INFO 0x10

#define OBJECT_NAME_NOT_VALID 0
#define OBJECT_NAME_VALID 0x11

      
#define AUTHENTIC_KEY_SIZE          0x80
#define AUTHENTIC_INFOBJ_LEN        0x0080  /* 128 bytes */
#define AUTHENTIC_MAXMSC_KEY_NUM    0x000F  /* 8 public, 8 private 0-5 */
#define AUTHENTIC_MAX_KEYS          0x0010

/*
#define MSC_SIZEOF_KEYINFO          11
#define MSC_SIZEOF_KEYNUMBER        1
#define MSC_SIZEOF_KEYPARTNER       1
#define MSC_SIZEOF_KEYSIZE          2
#define MSC_SIZEOF_KEYTYPE          1
#define MSC_SIZEOF_ACLVALUE         2
*/

#define MSC_EXP_SIZE    0x10
#define MSC_MOD_SIZE    0x80

#define OFFSET_ENCODING 0x00
#define OFFSET_KEYTYPE  0x01
#define OFFSET_KEYSIZE  0x02

#define CF_1024_KEYSIZE      1024
#define CF_1024P_MODSIZE      128
#define CF_1024P_EXPSIZE        4

#define MC_1024P_MOD            4
#define MC_1024P_EXP          134
#define MC_1024P_FULLSIZE     140

#ifdef __cplusplus
  }
#endif

#endif /* __authentIC_h__ */

