/******************************************************************
 
        MUSCLE SmartCard Development ( http://www.linuxnet.com )
            Title  : musclecard.h
            Package: MuscleCard Framework
            Author : David Corcoran
            Date   : 11/28/01
            License: Copyright (C) 2001 David Corcoran
                     <corcoran@linuxnet.com>
            Purpose: This abstracts the MUSCLE Card Edge Inteface

	    You may not remove this header from this file
	    without prior permission from the author.
 
********************************************************************/

#ifndef __pl_musclecard_h__
#define __pl_musclecard_h__

#ifdef __cplusplus
extern "C"
{
#endif

#ifdef __APPLE__
#include <PCSC/musclecard.h>
#else
#include <musclecard.h>
#endif

	/*
	 * Pre-personalization function 
	 */
#ifdef WIN32
	PCSC_API
#endif
	MSC_RV PL_MSCWriteFramework(MSCLPTokenConnection pConnection,
		MSCLPInitTokenParams pInitParams);

	/*****************************************************************/
	/*
	 * Core Musclecard functions 
	 */
	/*
	 * These functions coorespond directly to internal library 
	 */
	/*
	 * functions.  
	 */
	/*****************************************************************/

#ifdef WIN32
	PCSC_API
#endif
	MSC_RV PL_MSCGetStatus(MSCLPTokenConnection pConnection,
		MSCLPStatusInfo pStatusInfo);

#ifdef WIN32
	PCSC_API
#endif
	MSC_RV PL_MSCGetCapabilities(MSCLPTokenConnection pConnection,
		MSCULong32 Tag, MSCPUChar8 Value, MSCPULong32 Length);

#ifdef WIN32
	PCSC_API
#endif
	MSC_RV PL_MSCExtendedFeature(MSCLPTokenConnection pConnection,
		MSCULong32 extFeature,
		MSCPUChar8 outData,
		MSCULong32 outLength, MSCPUChar8 inData, MSCPULong32 inLength);

#ifdef WIN32
	PCSC_API
#endif
	MSC_RV PL_MSCGenerateKeys(MSCLPTokenConnection pConnection,
		MSCUChar8 prvKeyNum,
		MSCUChar8 pubKeyNum, MSCLPGenKeyParams pParams);

#ifdef WIN32
	PCSC_API
#endif
	MSC_RV PL_MSCImportKey(MSCLPTokenConnection pConnection,
		MSCUChar8 keyNum,
		MSCLPKeyACL pKeyACL,
		MSCPUChar8 pKeyBlob,
		MSCULong32 keyBlobSize,
		MSCLPKeyPolicy keyPolicy,
		MSCPVoid32 pAddParams, MSCUChar8 addParamsSize);

#ifdef WIN32
	PCSC_API
#endif
	MSC_RV PL_MSCExportKey(MSCLPTokenConnection pConnection,
		MSCUChar8 keyNum,
		MSCPUChar8 pKeyBlob,
		MSCPULong32 keyBlobSize,
		MSCPVoid32 pAddParams, MSCUChar8 addParamsSize);

#ifdef WIN32
	PCSC_API
#endif
	MSC_RV PL_MSCComputeCrypt(MSCLPTokenConnection pConnection,
		MSCLPCryptInit cryptInit,
		MSCPUChar8 pInputData,
		MSCULong32 inputDataSize,
		MSCPUChar8 pOutputData, MSCPULong32 outputDataSize);

#ifdef WIN32
	PCSC_API
#endif
	MSC_RV PL_MSCExtAuthenticate(MSCLPTokenConnection pConnection,
				  MSCUChar8 keyNum,
				  MSCUChar8 cipherMode,
				  MSCUChar8 cipherDirection, 
				  MSCPUChar8 pData, 
				  MSCULong32 dataSize);

#ifdef WIN32
	PCSC_API
#endif
	MSC_RV PL_MSCListKeys(MSCLPTokenConnection pConnection,
			   MSCUChar8 seqOption, 
			   MSCLPKeyInfo pKeyInfo);

#ifdef WIN32
	PCSC_API
#endif
	MSC_RV PL_MSCCreatePIN(MSCLPTokenConnection pConnection,
			    MSCUChar8 pinNum,
			    MSCUChar8 pinAttempts,
			    MSCPUChar8 pPinCode,
			    MSCULong32 pinCodeSize,
			    MSCPUChar8 pUnblockCode, 
			    MSCUChar8 unblockCodeSize);

#ifdef WIN32
	PCSC_API
#endif
	MSC_RV PL_MSCVerifyPIN(MSCLPTokenConnection pConnection,
			    MSCUChar8 pinNum, 
			    MSCPUChar8 pPinCode, 
			    MSCULong32 pinCodeSize);

#ifdef WIN32
	PCSC_API
#endif
	MSC_RV PL_MSCChangePIN(MSCLPTokenConnection pConnection,
		MSCUChar8 pinNum,
		MSCPUChar8 pOldPinCode,
		MSCUChar8 oldPinCodeSize,
		MSCPUChar8 pNewPinCode, MSCUChar8 newPinCodeSize);

#ifdef WIN32
	PCSC_API
#endif
	MSC_RV PL_MSCUnblockPIN(MSCLPTokenConnection pConnection,
		MSCUChar8 pinNum,
		MSCPUChar8 pUnblockCode, MSCULong32 unblockCodeSize);

#ifdef WIN32
	PCSC_API
#endif
	MSC_RV PL_MSCListPINs(MSCLPTokenConnection pConnection,
		MSCPUShort16 pPinBitMask);

#ifdef WIN32
	PCSC_API
#endif
	MSC_RV PL_MSCCreateObject(MSCLPTokenConnection pConnection,
		MSCString objectID,
		MSCULong32 objectSize, MSCLPObjectACL pObjectACL);

#ifdef WIN32
	PCSC_API
#endif
	MSC_RV PL_MSCDeleteObject(MSCLPTokenConnection pConnection,
		MSCString objectID, MSCUChar8 zeroFlag);

#ifdef WIN32
	PCSC_API
#endif
	MSC_RV PL_MSCWriteObject(MSCLPTokenConnection pConnection,
		MSCString objectID, MSCULong32 offset, 
		MSCPUChar8 pInputData, MSCUChar8 dataSize);

#ifdef WIN32
	PCSC_API
#endif
	MSC_RV PL_MSCReadObject(MSCLPTokenConnection pConnection,
		MSCString objectID, MSCULong32 offset, 
                MSCPUChar8 pOutputData, MSCUChar8 dataSize);

#ifdef WIN32
	PCSC_API
#endif
	MSC_RV PL_MSCReadAllocateObject(MSCLPTokenConnection pConnection,
		MSCString objectID, MSCPUChar8 *pOutputData, 
                MSCPULong32 dataSize,
                LPRWEventCallback rwCallback, MSCPVoid32 addParams);

#ifdef WIN32
	PCSC_API
#endif
	MSC_RV PL_MSCListObjects(MSCLPTokenConnection pConnection,
		MSCUChar8 seqOption, MSCLPObjectInfo pObjectInfo);

#ifdef WIN32
	PCSC_API
#endif
	MSC_RV PL_MSCLogoutAll(MSCLPTokenConnection pConnection);

#ifdef WIN32
	PCSC_API
#endif
	MSC_RV PL_MSCGetChallenge(MSCLPTokenConnection pConnection,
		MSCPUChar8 pSeed,
		MSCUShort16 seedSize,
		MSCPUChar8 pRandomData, MSCUShort16 randomDataSize);


#ifdef WIN32
	PCSC_API
#endif
    MSC_RV PL_MSCGetObjectAttributes( MSCLPTokenConnection pConnection,
			       MSCString objectID, MSCLPObjectInfo pObjectInfo );
    

#ifdef __cplusplus
}
#endif

#endif							/* __pl_musclecard_h__ */
